import os
import sys
from typing import NamedTuple
from typing import Optional

from flask import abort
from flask import Flask
from flask import request
from werkzeug.security import safe_join
from werkzeug.utils import append_slash_redirect

from lektor.admin.modules import register_modules
from lektor.builder import Builder
from lektor.buildfailures import FailureController
from lektor.db import Database
from lektor.db import Record
from lektor.reporter import CliReporter


class ResolveResult(NamedTuple):
    artifact_name: Optional[str]
    filename: Optional[str]
    record_path: Optional[str]
    alt: Optional[str]


class LektorInfo:
    def __init__(self, env, output_path, ui_lang="en", extra_flags=None, verbosity=0):
        self.env = env
        self.ui_lang = ui_lang
        self.output_path = output_path
        self.extra_flags = extra_flags
        self.verbosity = verbosity

    def get_pad(self):
        return Database(self.env).new_pad()

    def get_builder(self, pad=None):
        if pad is None:
            pad = self.get_pad()
        return Builder(pad, self.output_path, extra_flags=self.extra_flags)

    def get_failure_controller(self, pad=None):
        if pad is None:
            pad = self.get_pad()
        return FailureController(pad, self.output_path)

    def resolve_artifact(self, url_path, pad=None, redirect_slash=True):
        """Resolves an artifact and also triggers a build if necessary.
        Returns a tuple in the form ``(artifact_name, filename)`` where
        `artifact_name` can be `None` in case a file was targeted explicitly.
        """
        if pad is None:
            pad = self.get_pad()

        artifact_name = filename = record_path = alt = None

        # We start with trying to resolve a source and then use the
        # primary
        source = pad.resolve_url_path(url_path)
        if source is not None:
            # If the request path does not end with a slash but we
            # requested a URL that actually wants a trailing slash, we
            # append it.  This is consistent with what apache and nginx do
            # and it ensures our relative urls work.
            if (
                not url_path.endswith("/")
                and source.url_path != "/"
                and source.url_path != url_path
            ):
                return abort(append_slash_redirect(request.environ))

            with CliReporter(self.env, verbosity=self.verbosity):
                builder = self.get_builder(pad)
                prog, _ = builder.build(source)

            artifact = prog.primary_artifact
            if artifact is not None:
                artifact_name = artifact.artifact_name
                filename = artifact.dst_filename
            alt = source.alt
            if isinstance(source, Record):
                record_path = source.record.path

        if filename is None:
            path_list = url_path.strip("/").split("/")
            if sys.platform == "win32":
                filename = os.path.join(self.output_path, *path_list)
            else:
                filename = safe_join(self.output_path, *path_list)

        return ResolveResult(artifact_name, filename, record_path, alt)


class WebUI(Flask):
    def __init__(
        self,
        env,
        debug=False,
        output_path=None,
        ui_lang="en",
        verbosity=0,
        extra_flags=None,
    ):
        Flask.__init__(self, "lektor.admin", static_url_path="/admin/static")
        self.lektor_info = LektorInfo(
            env, output_path, ui_lang, extra_flags=extra_flags, verbosity=verbosity
        )
        self.debug = debug
        self.config["PROPAGATE_EXCEPTIONS"] = True

        register_modules(self)


WebAdmin = WebUI



import os
from typing import NamedTuple

import pytest

from lektor.quickstart import get_default_author
from lektor.quickstart import get_default_author_email
from lektor.utils import locate_executable


class struct_passwd(NamedTuple):
    pw_name: str = "user"
    pw_passwd: str = "pw"
    pw_uid: int = 10000
    pw_gid: int = 10000
    pw_gecos: str = "gecos"
    pw_dir: str = "/tmp"
    pw_shell: str = "/bin/false"


@pytest.fixture
def no_utils(monkeypatch):
    """Monkeypatch $PATH to hide any installed external utilities (e.g. git)."""
    monkeypatch.setitem(os.environ, "PATH", "/dev/null")
    locate_executable.cache_clear()


@pytest.fixture
def git_config_file(tmp_path, monkeypatch):
    """Create a temporary git config file, and monkeypatch $GIT_CONFIG to point to it."""
    config_file = tmp_path / "git_config"
    config_file.touch()
    monkeypatch.setitem(os.environ, "GIT_CONFIG", str(config_file))
    return config_file


@pytest.mark.skipif(os.name == "nt", reason="windows")
def test_default_author_from_pwd(mocker):
    pw_gecos = "Lektor Tester,,555-1212,,"
    mocker.patch(
        "pwd.getpwuid", spec=True, return_value=struct_passwd(pw_gecos=pw_gecos)
    )
    assert get_default_author() == "Lektor Tester"


def test_default_author_from_username(mocker):
    mocker.patch("getpass.getuser", spec=True, return_value="lektortester")
    if os.name != "nt":
        mocker.patch("os.getuid", spec=True, return_value=-1)
    assert get_default_author() == "lektortester"


@pytest.mark.skipif(locate_executable("git") is None, reason="git not installed")
def test_default_author_email(git_config_file):
    git_config_file.write_text("[user]\n\temail = tester@example.com\n")
    assert get_default_author_email() == "tester@example.com"


@pytest.mark.usefixtures("no_utils")
def test_default_author_email_from_EMAIL(monkeypatch):
    email = "tester@example.net"
    monkeypatch.setitem(os.environ, "EMAIL", email)
    assert get_default_author_email() == email


@pytest.mark.usefixtures("no_utils")
def test_default_author_email_no_default(monkeypatch):
    monkeypatch.delitem(os.environ, "EMAIL", raising=False)
    assert get_default_author_email() is None



import re
from typing import Union

import pytest
from markupsafe import Markup

from lektor.context import Context
from lektor.markdown import _markdown_cache
from lektor.markdown import ImprovedRenderer
from lektor.markdown import Markdown
from lektor.markdown import markdown_to_html
from lektor.pluginsystem import Plugin


def _xfail_966(*args):
    return pytest.param(
        *args,
        marks=pytest.mark.xfail(
            reason="Links are not currently resolved to lektor objects. See #966."
        ),
    )


@pytest.fixture
def record_path():
    return "/extra"


@pytest.fixture
def record_alt():
    return "en"


@pytest.fixture
def record(pad, record_path, record_alt):
    return pad.get(record_path, alt=record_alt)


@pytest.fixture
def base_url(record):
    return None


@pytest.fixture
def context(record, base_url):
    if base_url is None:
        base_url = record.url_path
    with Context(pad=record.pad) as ctx:
        with ctx.changed_base_url(base_url):
            yield ctx


@pytest.fixture
def renderer_context(context, record):
    ImprovedRenderer.record = record
    ImprovedRenderer.meta = {}
    try:
        yield
    finally:
        del ImprovedRenderer.record
        del ImprovedRenderer.meta


def _normalize_html(output: Union[str, Markup]) -> str:
    html = str(output).strip()
    html = html.replace("&copy;", "©")
    html = re.sub(r"(<img [^>]*?)\s*/>", r"\1>", html)
    return html


@pytest.mark.parametrize(
    "link, title, text, expected",
    [
        _xfail_966("a", None, "text", r'<a href="a/">text</a>\Z'),
        ("missing", None, "text", r'<a href="missing">text</a>\Z'),
        ("/", "T", "text", r'<a href="../" title="T">text</a>\Z'),
        ("a&amp;b", None, "x", r'.* href="a&amp;b"'),
        ("/", "<title>", "x", r'.* title="&lt;title&gt;"'),
    ],
)
@pytest.mark.usefixtures("renderer_context")
def test_ImprovedRenderer_link(link, title, text, expected):
    renderer = ImprovedRenderer()
    result = renderer.link(link, title, text)
    assert re.match(expected, result)


@pytest.mark.parametrize(
    "src, title, alt, expected",
    [
        ("/test.jpg", None, "text", r'<img src="../test.jpg" alt="text"\s*/?>\Z'),
        ("/test.jpg", "T", "x", r'.* title="T"'),
        ("&amp;c.gif", None, "x", r'.* src="&amp;c.gif"'),
        ("/test.jpg", "<title>", "x", r'.* title="&lt;title&gt;"'),
        ("/test.jpg", None, "x&y", r'.* alt="x&amp;y"'),
    ],
)
@pytest.mark.usefixtures("renderer_context")
def test_ImprovedRenderer_image(src, title, alt, expected):
    renderer = ImprovedRenderer()
    result = renderer.image(src, title, alt)
    assert re.match(expected, result.rstrip())


@pytest.mark.usefixtures("context")
def test_markdown_to_html(record):
    # pylint: disable=unused-variable
    html, meta = markdown_to_html("goober", record)
    assert _normalize_html(html) == "<p>goober</p>"


def test_markdown_to_html_requires_context(record):
    with pytest.raises(RuntimeError):
        markdown_to_html("goober", record)


class LinkCounterPlugin(Plugin):
    """A test plugin to test the renderer meta system."""

    name = "LinkCounter"

    class RendererMixin:
        def link(self, *args, **kwargs):
            self.meta["nlinks"] += 1
            return super().link(*args, **kwargs)

    def on_markdown_config(self, config, **kwargs):
        config.renderer_mixins.append(self.RendererMixin)

    def on_markdown_meta_init(self, meta, **kwargs):
        # pylint: disable=no-self-use
        meta["nlinks"] = 0


@pytest.fixture
def link_counter_plugin(env):
    env.plugin_controller.instanciate_plugin("link-counter", LinkCounterPlugin)


@pytest.fixture
def fresh_markdown_parser(env):
    """Disable Lektor's thread-local caching of Markdown parser."""
    cached = getattr(_markdown_cache, "md", None)
    if cached is not None:
        del _markdown_cache.md
    try:
        yield
    finally:
        if cached is not None:
            _markdown_cache.md = cached


@pytest.mark.parametrize(
    "source, nlinks",
    [
        ("howdy", 0),
        ("[howdy](http://example.com/)", 1),
    ],
)
@pytest.mark.usefixtures("context", "link_counter_plugin", "fresh_markdown_parser")
def test_markdown_to_html_meta(record, source, nlinks):
    # pylint: disable=unused-variable
    html, meta = markdown_to_html(source, record)
    assert meta["nlinks"] == nlinks


class TestMarkdown:
    # pylint: disable=no-self-use

    @pytest.fixture
    def source(self):
        return "text"

    @pytest.fixture
    def markdown(self, source, record):
        return Markdown(source, record)

    @pytest.mark.parametrize("source, expected", [("x", True), ("", False)])
    def test_bool(self, markdown, source, expected):
        assert bool(markdown) is expected

    @pytest.mark.usefixtures("context", "link_counter_plugin")
    def test_meta(self, markdown):
        assert markdown.meta["nlinks"] == 0

    @pytest.mark.usefixtures("context")
    def test_html(self, markdown):
        assert markdown.html.rstrip() == "<p>text</p>"

    @pytest.mark.usefixtures("context", "link_counter_plugin")
    def test_getitem(self, markdown):
        assert markdown["nlinks"] == 0

    @pytest.mark.usefixtures("context")
    def test_str(self, markdown):
        assert str(markdown).rstrip() == "<p>text</p>"

    @pytest.mark.usefixtures("context")
    def test_markup(self, markdown):
        assert markdown.__html__().rstrip() == "<p>text</p>"


@pytest.mark.parametrize(
    "source, expected",
    [
        ("text", r"<p>text</p>\Z"),
        ("&copy;", r"<p>©</p>\Z"),
        # Various link texts
        ("[text](x)", r".*<a [^>]*>text</a>"),
        ("[&copy;](x)", r".*<a [^>]*>©</a>"),
        ("[a&b](x)", r".*<a [^>]*>a&amp;b</a>"),
        ("[a&amp;b](x)", r".*<a [^>]*>a&amp;b</a>"),
        ("[a<>b](x)", r".*<a [^>]*>a&lt;&gt;b</a>"),
        ("[a<br>b](x)", r".*<a [^>]*>a<br>b</a>"),
        # Various link titles
        ('[x](x "TITLE")', r'.*<a\b.* title="TITLE">'),
        ('[x](x "©")', r'.*<a\b.* title="©">'),
        ('[x](x "&copy;")', r'.*<a\b.* title="©">'),
        ('[x](x "a&b")', r'.*<a\b.* title="a&amp;b">'),
        ('[x](x "a&amp;b")', r'.*<a\b.* title="a&amp;b">'),
        ('[x](x "a<br>b")', r'.*<a\b.* title="a&lt;br&gt;b">'),
        # Various link hrefs
        ("[x](&amp;c.html)", r'.*<a\b.* href="&amp;c.html">'),
        ("[x](&c.html)", r'.*<a\b.* href="&amp;c.html">'),
        # Various image alts
        ("![©](x)", r'.*<img\b[^>]* alt="©"'),
        ("![&copy;](x)", r'.*<img\b[^>]* alt="©"'),
        ("![<br>](x)", r'.*<img\b[^>]* alt="&lt;br&gt;"'),
        ("![&](x)", r'.*<img\b[^>]* alt="&amp;"'),
        ("![&amp;](x)", r'.*<img\b[^>]* alt="&amp;"'),
        # Various image titles
        ('![x](y "<br>")', r'.*<img\b[^>]* title="&lt;br&gt;"'),
        ('![x](y "&lt;")', r'.*<img\b[^>]* title="&lt;"'),
        # Various image srcs
        ("![](x&y.gif)", r'.*<img\b[^>]* src="x&amp;y.gif"'),
        ("![](x&amp;y.gif)", r'.*<img\b[^>]* src="x&amp;y.gif"'),
        ("![](x<br>y.gif)", r'.*<img\b[^>]* src="x&lt;br&gt;y.gif"'),
    ],
)
@pytest.mark.usefixtures("context")
def test_integration(record, source, expected):
    rendered = _normalize_html(Markdown(source, record))
    assert re.match(expected, rendered)



import os
import shutil
import sys
import textwrap

import pkg_resources
import pytest
from _pytest.monkeypatch import MonkeyPatch

import lektor.project
from lektor.builder import Builder
from lektor.db import Database
from lektor.db import Tree
from lektor.environment import Environment
from lektor.environment.expressions import Expression
from lektor.project import Project
from lektor.reporter import BufferReporter


@pytest.fixture(scope="session", autouse=True)
def temporary_lektor_cache(tmp_path_factory):
    """Get Lektor to use a temporary cache directory.

    This prevents the tests from leaving scats behind in the
    user’s real cache directory.

    """
    cache_dir = tmp_path_factory.mktemp("lektor_cache")

    # The stock monkeypatch fixture is function-scoped and so can not
    # be used in a session-scoped fixture.
    # Workaround from:
    # https://github.com/pytest-dev/pytest/issues/363#issuecomment-406536200

    def get_cache_dir():
        return str(cache_dir)

    mp = MonkeyPatch()
    mp.setattr(lektor.project, "get_cache_dir", get_cache_dir)
    yield cache_dir
    mp.undo()


@pytest.fixture
def save_sys_path(monkeypatch):
    """Save `sys.path`, `sys.modules`, and `pkg_resources` state on test
    entry, restore after test completion.

    Any test which constructs a `lektor.environment.Environment` instance
    or which runs any of the Lektor CLI commands should use this fixture
    to ensure that alternations made to `sys.path` do not interfere with
    other tests.

    Lektor's private package cache is added to `sys.path` by
    `lektor.packages.load_packages`.  This happens, for example,
    whenever a Lektor `Environment` is constructed (unless
    `load_plugins=False` is specified.)  Since all tests are run
    within an single invocation of the python interpreter, this can
    cause problems when different tests are using different private
    package caches.

    """
    monkeypatch.setattr(sys, "path", sys.path.copy())

    # Restoring `sys.modules` is an attempt to unload any
    # modules loaded during the test so that they can be re-loaded for
    # the next test.  This is not guaranteed to work, since there are
    # numerous ways that a reference to a loaded module may still be held.
    monkeypatch.setattr(sys, "modules", sys.modules.copy())

    # While pkg_resources.__getstate__ and pkg_resources.__setstate__
    # do not appear to be a documented part of the pkg_resources API,
    # they are used in setuptools' own tests, and appear to have been
    # a stable feature since 2011.
    saved_state = pkg_resources.__getstate__()
    yield
    pkg_resources.__setstate__(saved_state)


@pytest.fixture(scope="function")
def project():
    return Project.from_path(os.path.join(os.path.dirname(__file__), "demo-project"))


@pytest.fixture(scope="function")
def scratch_project_data(tmpdir):
    base = tmpdir.mkdir("scratch-proj")
    lektorfile_text = textwrap.dedent(
        """
        [project]
        name = Scratch

        [alternatives.en]
        primary = yes
        [alternatives.de]
        url_prefix = /de/
    """
    )
    base.join("Scratch.lektorproject").write_text(lektorfile_text, "utf8", ensure=True)
    content_text = textwrap.dedent(
        """
        _model: page
        ---
        title: Index
        ---
        body: *Hello World!*
    """
    )
    base.join("content", "contents.lr").write_text(content_text, "utf8", ensure=True)
    template_text = textwrap.dedent(
        """
        <h1>{{ this.title }}</h1>
        {{ this.body }}
    """
    )
    base.join("templates", "page.html").write_text(template_text, "utf8", ensure=True)
    model_text = textwrap.dedent(
        """
        [model]
        label = {{ this.title }}

        [fields.title]
        type = string
        [fields.body]
        type = markdown
    """
    )
    base.join("models", "page.ini").write_text(model_text, "utf8", ensure=True)

    return base


@pytest.fixture(scope="function")
def scratch_project(scratch_project_data):
    base = scratch_project_data
    return Project.from_path(str(base))


@pytest.fixture(scope="function")
def env(project, save_sys_path):
    return Environment(project)


@pytest.fixture(scope="function")
def scratch_env(scratch_project, save_sys_path):
    return Environment(scratch_project)


@pytest.fixture(scope="function")
def pad(env):
    return Database(env).new_pad()


@pytest.fixture(scope="function")
def scratch_pad(scratch_env):
    return Database(scratch_env).new_pad()


@pytest.fixture(scope="function")
def scratch_tree(scratch_pad):
    return Tree(scratch_pad)


@pytest.fixture(scope="function")
def builder(tmpdir, pad):
    return Builder(pad, str(tmpdir.mkdir("output")))


@pytest.fixture(scope="function")
def scratch_builder(tmpdir, scratch_pad):
    return Builder(scratch_pad, str(tmpdir.mkdir("output")))


# Builder for child-sources-test-project, a project to test that child sources
# are built even if they're filtered out by a pagination query.
@pytest.fixture(scope="function")
def child_sources_test_project_builder(tmpdir):
    project = Project.from_path(
        os.path.join(os.path.dirname(__file__), "child-sources-test-project")
    )
    env = Environment(project)
    pad = Database(env).new_pad()

    return Builder(pad, str(tmpdir.mkdir("output")))


@pytest.fixture(scope="function")
def eval_expr(env):
    def eval_expr(expr, **kwargs):
        expr = Expression(env, expr)
        return expr.evaluate(**kwargs)

    return eval_expr


@pytest.fixture(scope="function")
def reporter(request, env):
    reporter = BufferReporter(env)
    reporter.push()
    request.addfinalizer(reporter.pop)
    return reporter


@pytest.fixture(scope="function")
def project_cli_runner(isolated_cli_runner, project, save_sys_path):
    """
    Copy the project files into the isolated file system used by the
    Click test runner.
    """
    for entry in os.listdir(project.tree):
        entry_path = os.path.join(project.tree, entry)
        if os.path.isdir(entry_path):
            shutil.copytree(entry_path, entry)
        else:
            shutil.copy2(entry_path, entry)
    return isolated_cli_runner



import os
import posixpath

import click
from flask import Blueprint
from flask import current_app
from flask import g
from flask import jsonify
from flask import request

from lektor.admin.utils import eventstream
from lektor.constants import PRIMARY_ALT
from lektor.publisher import publish
from lektor.publisher import PublishError
from lektor.utils import is_valid_id


bp = Blueprint("api", __name__, url_prefix="/admin/api")


def get_record_and_parent(path):
    pad = g.admin_context.pad
    record = pad.get(path)
    if record is None:
        parent = pad.get(posixpath.dirname(path))
    else:
        parent = record.parent
    return record, parent


@bp.route("/pathinfo")
def get_path_info():
    """Returns the path segment information for a record."""
    alt = request.args.get("alt") or PRIMARY_ALT
    tree_item = g.admin_context.tree.get(request.args["path"])
    segments = []

    while tree_item is not None:
        segments.append(
            {
                "id": tree_item.id,
                "path": tree_item.path,
                "label_i18n": tree_item.get_record_label_i18n(alt),
                "exists": tree_item.exists,
                "can_have_children": tree_item.can_have_children,
            }
        )
        tree_item = tree_item.get_parent()

    segments.reverse()
    return jsonify(segments=segments)


@bp.route("/recordinfo")
def get_record_info():
    request_path = request.args["path"]
    alt = request.args.get("alt") or PRIMARY_ALT
    tree_item = g.admin_context.tree.get(request_path)

    return jsonify(
        id=tree_item.id,
        path=tree_item.path,
        label_i18n=tree_item.get_record_label_i18n(alt),
        exists=tree_item.exists,
        is_attachment=tree_item.is_attachment,
        attachments=[
            {
                "id": x.id,
                "path": x.path,
                "type": x.attachment_type,
            }
            for x in tree_item.iter_attachments()
        ],
        children=[
            {
                "id": x.id,
                "path": x.path,
                "label": x.id,
                "label_i18n": x.get_record_label_i18n(alt),
                "visible": x.is_visible,
            }
            for x in tree_item.iter_subpages()
        ],
        alts=[
            {
                "alt": _.id,
                "is_primary": _.id == PRIMARY_ALT,
                "primary_overlay": _.is_primary_overlay,
                "name_i18n": _.name_i18n,
                "exists": _.exists,
            }
            for _ in tree_item.alts.values()
        ],
        can_have_children=tree_item.can_have_children,
        can_have_attachments=tree_item.can_have_attachments,
        can_be_deleted=tree_item.can_be_deleted,
    )


@bp.route("/previewinfo")
def get_preview_info():
    alt = request.args.get("alt") or PRIMARY_ALT
    record = g.admin_context.pad.get(request.args["path"], alt=alt)
    if record is None:
        return jsonify(exists=False, url=None, is_hidden=True)
    return jsonify(exists=True, url=record.url_path, is_hidden=record.is_hidden)


@bp.route("/find", methods=["POST"])
def find():
    alt = request.values.get("alt") or PRIMARY_ALT
    lang = request.values.get("lang") or g.admin_context.info.ui_lang
    q = request.values.get("q")
    builder = current_app.lektor_info.get_builder()
    return jsonify(results=builder.find_files(q, alt=alt, lang=lang))


@bp.route("/browsefs", methods=["POST"])
def browsefs():
    alt = request.values.get("alt") or PRIMARY_ALT
    record = g.admin_context.pad.get(request.values["path"], alt=alt)
    okay = False
    if record is not None:
        if record.is_attachment:
            fn = record.attachment_filename
        else:
            fn = record.source_filename
        if os.path.exists(fn):
            click.launch(fn, locate=True)
            okay = True
    return jsonify(okay=okay)


@bp.route("/matchurl")
def match_url():
    record = g.admin_context.pad.resolve_url_path(
        request.args["url_path"], alt_fallback=False
    )
    if record is None:
        return jsonify(exists=False, path=None, alt=None)
    return jsonify(exists=True, path=record["_path"], alt=record["_alt"])


@bp.route("/rawrecord")
def get_raw_record():
    alt = request.args.get("alt") or PRIMARY_ALT
    ts = g.admin_context.tree.edit(request.args["path"], alt=alt)
    return jsonify(ts.to_json())


@bp.route("/newrecord")
def get_new_record_info():
    pad = g.admin_context.pad
    alt = request.args.get("alt") or PRIMARY_ALT
    tree_item = g.admin_context.tree.get(request.args["path"])

    def describe_model(model):
        primary_field = None
        if model.primary_field is not None:
            f = model.field_map.get(model.primary_field)
            if f is not None:
                primary_field = f.to_json(pad)
        return {
            "id": model.id,
            "name": model.name,
            "name_i18n": model.name_i18n,
            "primary_field": primary_field,
        }

    implied_model = tree_item.implied_child_datamodel
    label_i18n = tree_item.get_record_label_i18n(alt)
    return jsonify(
        {
            "label_i18n": label_i18n,
            "label": label_i18n["en"],
            "can_have_children": tree_item.can_have_children,
            "implied_model": implied_model,
            "available_models": dict(
                (k, describe_model(v))
                for k, v in pad.db.datamodels.items()
                if not v.hidden or k == implied_model
            ),
        }
    )


@bp.route("/newattachment")
def get_new_attachment_info():
    alt = request.args.get("alt") or PRIMARY_ALT
    tree_item = g.admin_context.tree.get(request.args["path"])
    label_i18n = tree_item.get_record_label_i18n(alt)
    return jsonify(
        {
            "can_upload": tree_item.can_have_attachments,
            "label_i18n": label_i18n,
            "label": label_i18n["en"],
        }
    )


@bp.route("/newattachment", methods=["POST"])
def upload_new_attachments():
    alt = request.values.get("alt") or PRIMARY_ALT
    ts = g.admin_context.tree.edit(request.values["path"], alt=alt)
    if not ts.exists or ts.is_attachment:
        return jsonify({"bad_upload": True})

    buckets = []

    for file in request.files.getlist("file"):
        buckets.append(
            {
                "original_filename": file.filename,
                "stored_filename": ts.add_attachment(file.filename, file),
            }
        )

    return jsonify(
        {
            "bad_upload": False,
            "path": request.form["path"],
            "buckets": buckets,
        }
    )


@bp.route("/newrecord", methods=["POST"])
def add_new_record():
    values = request.get_json()
    alt = values.get("alt") or PRIMARY_ALT
    exists = False

    if not is_valid_id(values["id"]):
        return jsonify(valid_id=False, exists=False, path=None)

    path = posixpath.join(values["path"], values["id"])

    ts = g.admin_context.tree.edit(path, datamodel=values.get("model"), alt=alt)
    with ts:
        if ts.exists:
            exists = True
        else:
            ts.data.update(values.get("data") or {})

    return jsonify({"valid_id": True, "exists": exists, "path": path})


@bp.route("/deleterecord", methods=["POST"])
def delete_record():
    alt = request.values.get("alt") or PRIMARY_ALT
    delete_master = request.values.get("delete_master") == "1"
    if request.values["path"] != "/":
        ts = g.admin_context.tree.edit(request.values["path"], alt=alt)
        with ts:
            ts.delete(delete_master=delete_master)
    return jsonify(okay=True)


@bp.route("/rawrecord", methods=["PUT"])
def update_raw_record():
    values = request.get_json()
    data = values["data"]
    alt = values.get("alt") or PRIMARY_ALT
    ts = g.admin_context.tree.edit(values["path"], alt=alt)
    with ts:
        ts.data.update(data)
    return jsonify(path=ts.path)


@bp.route("/servers")
def get_servers():
    db = g.admin_context.pad.db
    config = db.env.load_config()
    servers = config.get_servers(public=True)
    return jsonify(
        servers=sorted(
            [x.to_json() for x in servers.values()], key=lambda x: x["name"].lower()
        )
    )


@bp.route("/build", methods=["POST"])
def trigger_build():
    builder = current_app.lektor_info.get_builder()
    builder.build_all()
    builder.prune()
    return jsonify(okay=True)


@bp.route("/clean", methods=["POST"])
def trigger_clean():
    builder = current_app.lektor_info.get_builder()
    builder.prune(all=True)
    builder.touch_site_config()
    return jsonify(okay=True)


@bp.route("/publish")
def publish_build():
    db = g.admin_context.pad.db
    server = request.values["server"]
    config = db.env.load_config()
    server_info = config.get_server(server)
    info = current_app.lektor_info

    @eventstream
    def generator():
        try:
            event_iter = (
                publish(
                    info.env,
                    server_info.target,
                    info.output_path,
                    server_info=server_info,
                )
                or ()
            )
            for event in event_iter:
                yield {"msg": event}
        except PublishError as e:
            yield {"msg": "Error: %s" % e}

    return generator()


@bp.route("/ping")
def ping():
    return jsonify(project_id=current_app.lektor_info.env.project.id, okay=True)



import json
import mimetypes
import os
import string
from io import BytesIO
from zlib import adler32

from flask import abort
from flask import Blueprint
from flask import current_app
from flask import render_template
from flask import request
from flask import Response
from flask import url_for
from werkzeug.exceptions import NotFound
from werkzeug.wsgi import wrap_file

from lektor.constants import PRIMARY_ALT

bp = Blueprint("serve", __name__)


_EDIT_BUTTON_STYLE = """
  <style type="text/css">
    #lektor-edit-link {
      position: fixed;
      z-index: 9999999;
      right: 10px;
      top: 10px;
      position: fixed;
      margin: 0;
      font-family: 'Verdana', sans-serif;
      background: #eee;
      color: #77304c;
      font-weight: normal;
      font-size: 32px;
      padding: 0;
      text-decoration: none!important;
      border: 1px solid #ccc!important;
      width: 40px;
      height: 40px;
      line-height: 40px;
      text-align: center;
      opacity: 0.7;
    }

    #lektor-edit-link:hover {
      background: white!important;
      opacity: 1.0;
      border: 1px solid #aaa!important;
    }
  </style>
"""


_EDIT_BUTTON_SCRIPT_TMPL = string.Template(
    """
  <script type="text/javascript">
    (function() {
      if (window != window.top) {
        return;
      }
      var link = document.createElement('a');
      link.setAttribute('href', ${edit_url});
      link.setAttribute('id', 'lektor-edit-link');
      link.innerHTML = '\u270E';
      document.body.appendChild(link);
    })();
  </script>
"""
)


def rewrite_html_for_editing(fp, edit_url):
    contents = fp.read()
    fp.close()
    head_endpos = contents.find(b"</head>")
    body_endpos = contents.find(b"</body>")
    if head_endpos < 0:
        # If </head> not found, inject both <style> and <script> at end
        head_endpos = body_endpos = len(contents)
    elif body_endpos < 0:
        # If </body> not found, inject <script> at end
        body_endpos = len(contents)

    button_script = _EDIT_BUTTON_SCRIPT_TMPL.substitute(
        edit_url=json.dumps(edit_url)
    ).encode("utf-8")

    return BytesIO(
        contents[:head_endpos]
        + _EDIT_BUTTON_STYLE.encode("utf-8")
        + contents[head_endpos:body_endpos]
        + button_script
        + contents[body_endpos:]
    )


def send_file(fp, mimetype):
    try:
        fileno = fp.fileno()
        stat = os.stat(fileno)
    except OSError:
        stat = None

    resp = Response(
        wrap_file(request.environ, fp),
        mimetype=mimetype,
        direct_passthrough=True,
    )
    resp.cache_control.no_store = True
    if stat is not None:
        resp.cache_control.public = True
        resp.content_length = stat.st_size
        check = adler32(f"{stat.st_dev}:{stat.st_ino}".encode("ascii")) & 0xFFFFFFFF
        resp.set_etag(f"lektor-{stat.st_mtime}-{stat.st_size}-{check}")
    return resp


def handle_build_failure(failure):
    return render_template("build-failure.html", **failure.data)


def serve_up_artifact(path):
    li = current_app.lektor_info
    pad = li.get_pad()

    resolved = li.resolve_artifact("/" + path, pad)
    if resolved.filename is None:
        abort(404)

    artifact_name = resolved.artifact_name
    if artifact_name is None:
        artifact_name = path.strip("/")

    # If there was a build failure for the given artifact, we want
    # to render this instead of sending the (most likely missing or
    # corrupted) file.
    ctrl = li.get_failure_controller(pad)
    failure = ctrl.lookup_failure(artifact_name)
    if failure is not None:
        return handle_build_failure(failure)

    mimetype = mimetypes.guess_type(resolved.filename)[0]
    if mimetype is None:
        mimetype = "application/octet-stream"

    try:
        # pylint: disable=consider-using-with
        fp = open(resolved.filename, "rb")
    except OSError:
        abort(404)

    if mimetype == "text/html" and resolved.record_path is not None:
        assert "@" not in resolved.record_path
        alt = resolved.alt
        if not alt or alt in (PRIMARY_ALT, pad.db.config.primary_alternative):
            alt = None
        edit_url = url_for("dash.app", view="edit", path=resolved.record_path, alt=alt)
        fp = rewrite_html_for_editing(fp, edit_url)

    return send_file(fp, mimetype)


@bp.route("/", defaults={"path": ""})
@bp.route("/<path:path>")
def serve_artifact(path):
    return serve_up_artifact(path)


@bp.errorhandler(404)
def serve_error_page(error):
    try:
        return serve_up_artifact("404.html")
    except NotFound as e:
        return e



import errno
import hashlib
import json
import os
from traceback import TracebackException


class BuildFailure:
    def __init__(self, data):
        self.data = data

    @classmethod
    def from_exc_info(cls, artifact_name, exc_info):
        te = TracebackException(*exc_info)
        # NB: we have dropped werkzeug's support for Paste's __traceback_hide__
        # frame local.
        return cls(
            {
                "artifact": artifact_name,
                "exception": "".join(te.format_exception_only()).strip(),
                "traceback": "".join(te.format()).strip(),
            }
        )

    def to_json(self):
        return self.data


class FailureController:
    def __init__(self, pad, destination_path):
        self.pad = pad
        self.path = os.path.join(
            os.path.abspath(os.path.join(pad.db.env.root_path, destination_path)),
            ".lektor",
            "failures",
        )

    def get_filename(self, artifact_name):
        return (
            os.path.join(
                self.path, hashlib.md5(artifact_name.encode("utf-8")).hexdigest()
            )
            + ".json"
        )

    def lookup_failure(self, artifact_name):
        """Looks up a failure for the given artifact name."""
        fn = self.get_filename(artifact_name)
        try:
            with open(fn, "r", encoding="utf-8") as f:
                return BuildFailure(json.load(f))
        except IOError as e:
            if e.errno != errno.ENOENT:
                raise
            return None

    def clear_failure(self, artifact_name):
        """Clears a stored failure."""
        try:
            os.unlink(self.get_filename(artifact_name))
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    def store_failure(self, artifact_name, exc_info):
        """Stores a failure from an exception info tuple."""
        fn = self.get_filename(artifact_name)
        try:
            os.makedirs(os.path.dirname(fn))
        except OSError:
            pass
        with open(fn, mode="w", encoding="utf-8") as f:
            json.dump(BuildFailure.from_exc_info(artifact_name, exc_info).to_json(), f)
            f.write("\n")



class LektorException(Exception):
    def __init__(self, message=None):
        Exception.__init__(self)
        if isinstance(message, bytes):
            message = message.decode("utf-8", "replace")
        self.message = message

    def to_json(self):
        return {
            "type": self.__class__.__name__,
            "message": self.message,
        }

    def __str__(self):
        return str(self.message)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.message)



import inspect
import os
import sys
import warnings
from weakref import ref as weakref

import pkg_resources
from inifile import IniFile
from pkg_resources import get_distribution

from lektor.context import get_ctx
from lektor.utils import process_extra_flags


def get_plugin(plugin_id_or_class, env=None):
    """Looks up the plugin instance by id or class."""
    if env is None:
        ctx = get_ctx()
        if ctx is None:
            raise RuntimeError(
                "Context is unavailable and no environment "
                "was passed to the function."
            )
        env = ctx.env
    plugin_id = env.plugin_ids_by_class.get(plugin_id_or_class, plugin_id_or_class)
    try:
        return env.plugins[plugin_id]
    except KeyError as error:
        raise LookupError("Plugin %r not found" % plugin_id) from error


class Plugin:
    """This needs to be subclassed for custom plugins."""

    name = "Your Plugin Name"
    description = "Description goes here"

    def __init__(self, env, id):
        self._env = weakref(env)
        self.id = id

    @property
    def env(self):
        rv = self._env()
        if rv is None:
            raise RuntimeError("Environment went away")
        return rv

    @property
    def version(self):

        return get_distribution("lektor-" + self.id).version

    @property
    def path(self):
        mod = sys.modules[self.__class__.__module__.split(".", maxsplit=1)[0]]
        path = os.path.abspath(os.path.dirname(mod.__file__))
        if not path.startswith(self.env.project.get_package_cache_path()):
            return path
        return None

    @property
    def import_name(self):
        return self.__class__.__module__ + ":" + self.__class__.__name__

    def get_lektor_config(self):
        """Returns the global config."""
        ctx = get_ctx()
        if ctx is not None:
            cfg = ctx.pad.db.config
        else:
            cfg = self.env.load_config()
        return cfg

    @property
    def config_filename(self):
        """The filename of the plugin specific config file."""
        return os.path.join(self.env.root_path, "configs", self.id + ".ini")

    def get_config(self, fresh=False):
        """Returns the config specific for this plugin.  By default this
        will be cached for the current build context but this can be
        disabled by passing ``fresh=True``.
        """
        ctx = get_ctx()
        if ctx is not None and not fresh:
            cache = ctx.cache.setdefault(__name__ + ":configs", {})
            cfg = cache.get(self.id)
            if cfg is None:
                cfg = IniFile(self.config_filename)
                cache[self.id] = cfg
        else:
            cfg = IniFile(self.config_filename)
        if ctx is not None:
            ctx.record_dependency(self.config_filename)
        return cfg

    def emit(self, event, **kwargs):
        return self.env.plugin_controller.emit(self.id + "-" + event, **kwargs)

    def to_json(self):
        return {
            "id": self.id,
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "path": self.path,
            "import_name": self.import_name,
        }


def load_plugins():
    """Loads all available plugins and returns them."""
    rv = {}
    for ep in pkg_resources.iter_entry_points("lektor.plugins"):
        match_name = "lektor-" + ep.name.lower()
        if match_name != ep.dist.project_name.lower():
            raise RuntimeError(
                "Disallowed distribution name: distribution name for "
                f"plugin {ep.name!r} must be {match_name!r}."
            )
        rv[ep.name] = ep.load()
    return rv


def initialize_plugins(env):
    """Initializes the plugins for the environment."""
    plugins = load_plugins()
    for plugin_id, plugin_cls in plugins.items():
        env.plugin_controller.instanciate_plugin(plugin_id, plugin_cls)
    env.plugin_controller.emit("setup-env")


class PluginController:
    """Helper management class that is used to control plugins through
    the environment.
    """

    def __init__(self, env, extra_flags=None):
        self._env = weakref(env)
        self.extra_flags = extra_flags

    @property
    def env(self):
        rv = self._env()
        if rv is None:
            raise RuntimeError("Environment went away")
        return rv

    def instanciate_plugin(self, plugin_id, plugin_cls):
        env = self.env
        if plugin_id in env.plugins:
            raise RuntimeError('Plugin "%s" is already registered' % plugin_id)
        env.plugins[plugin_id] = plugin_cls(env, plugin_id)
        env.plugin_ids_by_class[plugin_cls] = plugin_id

    def iter_plugins(self):
        # XXX: sort?
        return self.env.plugins.values()

    def emit(self, event, **kwargs):
        """Invoke event hook for all plugins that support it.

        Any ``kwargs`` are passed to the hook methods.

        Returns a dict mapping plugin ids to hook method return values.
        """
        rv = {}
        extra_flags = process_extra_flags(self.extra_flags)
        funcname = "on_" + event.replace("-", "_")
        for plugin in self.iter_plugins():
            handler = getattr(plugin, funcname, None)
            if handler is not None:
                kw = {**kwargs, "extra_flags": extra_flags}
                try:
                    inspect.signature(handler).bind(**kw)
                except TypeError:
                    del kw["extra_flags"]
                rv[plugin.id] = handler(**kw)
                if "extra_flags" not in kw:
                    warnings.warn(
                        f"The plugin {plugin.id!r} function {funcname!r} does not "
                        "accept extra_flags. "
                        "It should be updated to accept `**extra` so that it will "
                        "not break if new parameters are passed to it by newer "
                        "versions of Lektor.",
                        DeprecationWarning,
                    )
        return rv



from lektor.types.primitives import SingleInputType
from lektor.utils import slugify
from lektor.utils import Url


class SortKeyType(SingleInputType):
    widget = "integer"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing sort key")
        try:
            return int(raw.value.strip())
        except ValueError:
            return raw.bad_value("Bad sort key value")


class SlugType(SingleInputType):
    widget = "slug"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing slug")
        return slugify(raw.value)


class UrlType(SingleInputType):
    widget = "url"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing URL")
        return Url(raw.value)



"""Unit tests for lektor.pluginsystem.
"""
from pathlib import Path
from unittest import mock

import pkg_resources
import pytest

from lektor.cli import cli
from lektor.context import Context
from lektor.packages import add_package_to_project
from lektor.pluginsystem import get_plugin
from lektor.pluginsystem import load_plugins
from lektor.pluginsystem import Plugin
from lektor.pluginsystem import PluginController


class DummyPlugin(Plugin):
    name = "Dummy Plugin"
    description = "For testing"

    # DummyPlugin.calls is test-local (see fixture dummy_plugin_calls, below)
    calls = []

    def __getattr__(self, name):
        if not name.startswith("on_"):
            raise AttributeError(name)

        event = name[3:].replace("_", "-")

        def hook(extra_flags, **kwargs):
            call = {
                "event": event,
                "extra_flags": extra_flags,
                "kwargs": kwargs,
            }
            self.calls.append(call)
            return f"{event} return value"

        return hook

    def on_legacy_event(self):
        self.calls.append({"event": "legacy-event"})
        return "legacy-event return value"

    def on_one_type_error(self, **kwargs):
        """Raises TypeError only on the first call."""
        self.calls.append({"event": "one-type-error"})
        if len(self.calls) == 1:
            raise TypeError("test")
        return "one-type-error return value"


@pytest.fixture(autouse=True)
def dummy_plugin_calls(monkeypatch):
    """Reset list of DummyPlugin hook calls for each test."""
    monkeypatch.setattr(DummyPlugin, "calls", [])
    return DummyPlugin.calls


class DummyEntryPointMetadata:
    """Implement enough of `pkg_resources.IMetadataProvider` to convince a
    Distribution that it has an entry point.
    """

    # pylint: disable=no-self-use

    def __init__(self, entry_points_txt):
        self.entry_points_txt = entry_points_txt

    def has_metadata(self, name):
        return name == "entry_points.txt"

    def get_metadata(self, name):
        return self.entry_points_txt if name == "entry_points.txt" else ""

    def get_metadata_lines(self, name):
        return pkg_resources.yield_lines(self.get_metadata(name))


@pytest.fixture
def dummy_plugin_distribution_name():
    return "lektor-dummy-plugin"


@pytest.fixture
def dummy_plugin_distribution(dummy_plugin_distribution_name, save_sys_path):
    """Add a dummy plugin distribution to the current working_set."""
    dist = pkg_resources.Distribution(
        project_name=dummy_plugin_distribution_name,
        metadata=DummyEntryPointMetadata(
            f"""
            [lektor.plugins]
            dummy-plugin = {__name__}:DummyPlugin
            """
        ),
        version="1.23",
        location=__file__,
    )
    pkg_resources.working_set.add(dist)
    return dist


@pytest.fixture
def dummy_plugin(env):
    """Instantiate and register a dummy plugin in env"""
    env.plugin_controller.instanciate_plugin("dummy-plugin", DummyPlugin)
    return env.plugins["dummy-plugin"]


def test_get_plugin(env, dummy_plugin):
    assert get_plugin("dummy-plugin", env) == dummy_plugin


def test_get_plugin_from_context(env, dummy_plugin):
    with Context(pad=env.new_pad()):
        assert get_plugin("dummy-plugin") == dummy_plugin


def test_get_plugin_missing(env):
    with pytest.raises(LookupError, match=r"Plugin .* not found"):
        get_plugin("dummy-plugin", env)


def test_get_plugin_no_env_or_ctx(dummy_plugin):
    with pytest.raises(RuntimeError, match=r"Context is unavailable"):
        get_plugin("dummy-plugin")


class TestPlugin:
    # pylint: disable=no-self-use

    @pytest.fixture
    def scratch_project_data(self, scratch_project_data):
        """Add plugin config file to scratch project."""
        scratch_project_data.join("configs", "dummy-plugin.ini").write_text(
            "test_setting = test value\n",
            encoding="utf-8",
            ensure=True,
        )
        return scratch_project_data

    @pytest.fixture
    def scratch_plugin(self, scratch_env):
        """Instantiate and register a dummy plugin with scratch_env"""
        scratch_env.plugin_controller.instanciate_plugin("dummy-plugin", DummyPlugin)
        return scratch_env.plugins["dummy-plugin"]

    def test_env(self, dummy_plugin, env):
        assert dummy_plugin.env is env

    def test_env_went_away(self):
        env = mock.Mock(name="env", spec=())
        plugin = DummyPlugin(env, "dummy-plugin")
        del env
        with pytest.raises(RuntimeError, match=r"Environment went away"):
            getattr(plugin, "env")

    def test_version(self, dummy_plugin, dummy_plugin_distribution):
        assert dummy_plugin.version == dummy_plugin_distribution.version

    def test_path(self, dummy_plugin):
        assert dummy_plugin.path == str(Path(__file__).parent)

    @pytest.mark.requiresinternet
    @pytest.mark.usefixtures("save_sys_path")
    def test_path_installed_plugin_is_none(self, scratch_project):
        # XXX: this test is slow and fragile. (It won't run
        # without an internet connection.)
        add_package_to_project(scratch_project, "webpack-support")
        env = scratch_project.make_env()
        plugin = get_plugin("webpack-support", env)
        assert plugin.path is None

    def test_import_name(self, dummy_plugin):
        assert dummy_plugin.import_name == "test_pluginsystem:DummyPlugin"

    def test_get_lektor_config(self, dummy_plugin):
        cfg = dummy_plugin.get_lektor_config()
        assert cfg["PROJECT"]["name"] == "Demo Project"

    def test_get_lektor_config_from_context(self, dummy_plugin, env):
        with Context(pad=env.new_pad()):
            cfg = dummy_plugin.get_lektor_config()
        assert cfg["PROJECT"]["name"] == "Demo Project"

    def test_config_filename(self, dummy_plugin, env):
        configdir = Path(env.root_path) / "configs"
        assert dummy_plugin.config_filename == str(configdir / "dummy-plugin.ini")

    def test_get_config(self, scratch_plugin):
        cfg = scratch_plugin.get_config()
        assert cfg["test_setting"] == "test value"

    def test_get_config_records_dependency(self, scratch_plugin, scratch_env):
        with Context(pad=scratch_env.new_pad()) as ctx:
            scratch_plugin.get_config()
        assert scratch_plugin.config_filename in ctx.referenced_dependencies

    def test_get_config_returns_cached_value(self, scratch_plugin, scratch_env):
        with Context(pad=scratch_env.new_pad()):
            cfg = scratch_plugin.get_config()
            cfg["is_cached"] = "indeed"
            cfg = scratch_plugin.get_config()
        assert "is_cached" in cfg

    def test_get_config_fresh(self, scratch_plugin, scratch_env):
        with Context(pad=scratch_env.new_pad()):
            cfg = scratch_plugin.get_config()
            cfg["is_cached"] = "indeed"
            cfg = scratch_plugin.get_config(fresh=True)
        assert "is_cached" not in cfg

    def test_emit(self, dummy_plugin):
        rv = dummy_plugin.emit("subevent")
        assert rv == {"dummy-plugin": "dummy-plugin-subevent return value"}
        events = [call["event"] for call in dummy_plugin.calls]
        assert events == ["dummy-plugin-subevent"]

    def test_to_json(self, dummy_plugin, dummy_plugin_distribution):
        assert dummy_plugin.to_json() == {
            "id": "dummy-plugin",
            "name": DummyPlugin.name,
            "description": DummyPlugin.description,
            "version": dummy_plugin_distribution.version,
            "import_name": "test_pluginsystem:DummyPlugin",
            "path": str(Path(__file__).parent),
        }


def test_load_plugins(dummy_plugin_distribution):
    assert load_plugins() == {"dummy-plugin": DummyPlugin}


@pytest.mark.parametrize("dummy_plugin_distribution_name", ["evil-dist-name"])
def test_load_plugins_bad_distname(dummy_plugin_distribution):
    with pytest.raises(RuntimeError, match=r"Disallowed distribution name"):
        load_plugins()


class TestPluginController:
    # pylint: disable=no-self-use

    @pytest.fixture
    def extra_flags(self):
        return {"flag": "value"}

    @pytest.fixture
    def plugin_controller(self, env, extra_flags):
        return PluginController(env, extra_flags)

    def test_env(self, plugin_controller, env):
        assert plugin_controller.env is env

    def test_env_went_away(self):
        env = mock.Mock(name="env", spec=())
        plugin_controller = PluginController(env)
        del env
        with pytest.raises(RuntimeError, match=r"Environment went away"):
            getattr(plugin_controller, "env")

    def test_instantiate_plugin(self, plugin_controller, env):
        plugin_controller.instanciate_plugin("plugin-id", DummyPlugin)
        assert isinstance(env.plugins["plugin-id"], DummyPlugin)
        assert env.plugin_ids_by_class[DummyPlugin] == "plugin-id"

    def test_instantiate_plugin_error(self, plugin_controller, dummy_plugin):
        with pytest.raises(RuntimeError, match=r"already registered"):
            plugin_controller.instanciate_plugin(dummy_plugin.id, DummyPlugin)

    def test_iter_plugins(self, plugin_controller, dummy_plugin):
        assert list(plugin_controller.iter_plugins()) == [dummy_plugin]

    def test_emit(self, plugin_controller, dummy_plugin, extra_flags):
        rv = plugin_controller.emit("test-event")
        assert rv == {dummy_plugin.id: "test-event return value"}
        (call,) = dummy_plugin.calls
        assert call["extra_flags"] == extra_flags

    def test_emit_with_kwargs(self, plugin_controller, dummy_plugin):
        rv = plugin_controller.emit("another-test-event", another_arg="x")
        assert rv == {dummy_plugin.id: "another-test-event return value"}
        (call,) = dummy_plugin.calls
        assert call["kwargs"] == {"another_arg": "x"}

    def test_emit_deprecation_warning(self, plugin_controller, dummy_plugin):
        with pytest.deprecated_call():
            rv = plugin_controller.emit("legacy-event")
        assert rv == {dummy_plugin.id: "legacy-event return value"}

    def test_emit_is_not_confused_by_type_error(self, plugin_controller, dummy_plugin):
        # Excercises https://github.com/lektor/lektor/issues/1085
        with pytest.raises(TypeError):
            plugin_controller.emit("one-type-error")
        rv = plugin_controller.emit("one-type-error")
        assert rv == {dummy_plugin.id: "one-type-error return value"}


@pytest.mark.usefixtures("dummy_plugin_distribution")
def test_cli_integration(project, cli_runner, monkeypatch):
    """Check that plugin hooks receive extra_flags from command line."""
    # chdir to tests/demo-project
    monkeypatch.chdir(project.tree)
    result = cli_runner.invoke(
        cli, ["clean", "--yes", "-f", "flag1", "--extra-flag", "flag2:value2"]
    )
    print(result.output)
    assert result.exit_code == 0
    assert {"before-prune", "after-prune", "setup-env"}.issubset(
        call["event"] for call in DummyPlugin.calls
    )
    for call in DummyPlugin.calls:
        assert call["extra_flags"] == {"flag1": "flag1", "flag2": "value2"}



import time
import traceback
from contextlib import contextmanager

import click
from click import style
from werkzeug.local import LocalProxy
from werkzeug.local import LocalStack


_reporter_stack = LocalStack()
_build_buffer_stack = LocalStack()


def describe_build_func(func):
    self = getattr(func, "__self__", None)
    if self is not None and any(
        x.__name__ == "BuildProgram" for x in self.__class__.__mro__
    ):
        return self.__class__.__module__ + "." + self.__class__.__name__
    return func.__module__ + "." + func.__name__


class Reporter:
    def __init__(self, env, verbosity=0):
        self.env = env
        self.verbosity = verbosity

        self.builder_stack = []
        self.artifact_stack = []
        self.source_stack = []

    def push(self):
        _reporter_stack.push(self)

    @staticmethod
    def pop():
        _reporter_stack.pop()

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.pop()

    @property
    def builder(self):
        if self.builder_stack:
            return self.builder_stack[-1]
        return None

    @property
    def current_artifact(self):
        if self.artifact_stack:
            return self.artifact_stack[-1]
        return None

    @property
    def current_source(self):
        if self.source_stack:
            return self.source_stack[-1]
        return None

    @property
    def show_build_info(self):
        return self.verbosity >= 1

    @property
    def show_tracebacks(self):
        return self.verbosity >= 1

    @property
    def show_current_artifacts(self):
        return self.verbosity >= 2

    @property
    def show_artifact_internals(self):
        return self.verbosity >= 3

    @property
    def show_source_internals(self):
        return self.verbosity >= 3

    @property
    def show_debug_info(self):
        return self.verbosity >= 4

    @contextmanager
    def build(self, activity, builder):
        now = time.time()
        self.builder_stack.append(builder)
        self.start_build(activity)
        try:
            yield
        finally:
            self.builder_stack.pop()
            self.finish_build(activity, now)

    def start_build(self, activity):
        pass

    def finish_build(self, activity, start_time):
        pass

    @contextmanager
    def build_artifact(self, artifact, build_func, is_current):
        now = time.time()
        self.artifact_stack.append(artifact)
        self.start_artifact_build(is_current)
        self.report_build_func(build_func)
        try:
            yield
        finally:
            self.finish_artifact_build(now)
            self.artifact_stack.pop()

    def start_artifact_build(self, is_current):
        pass

    def finish_artifact_build(self, start_time):
        pass

    def report_failure(self, artifact, exc_info):
        pass

    def report_build_all_failure(self, failures):
        pass

    def report_dependencies(self, dependencies):
        for dep in dependencies:
            self.report_debug_info("dependency", dep[1])

    def report_dirty_flag(self, value):
        pass

    def report_write_source_info(self, info):
        pass

    def report_prune_source_info(self, source):
        pass

    def report_sub_artifact(self, artifact):
        pass

    def report_build_func(self, build_func):
        pass

    def report_debug_info(self, key, value):
        pass

    def report_generic(self, message):
        pass

    def report_pruned_artifact(self, artifact_name):
        pass

    @contextmanager
    def process_source(self, source):
        now = time.time()
        self.source_stack.append(source)
        self.enter_source()
        try:
            yield
        finally:
            self.leave_source(now)
            self.source_stack.pop()

    def enter_source(self):
        pass

    def leave_source(self, start_time):
        pass


class NullReporter(Reporter):
    pass


class BufferReporter(Reporter):
    def __init__(self, env, verbosity=0):
        Reporter.__init__(self, env, verbosity)
        self.buffer = []

    def clear(self):
        self.buffer = []

    def get_recorded_dependencies(self):
        rv = set()
        for event, data in self.buffer:
            if event == "debug-info" and data["key"] == "dependency":
                rv.add(data["value"])
        return sorted(rv)

    def get_major_events(self):
        rv = []
        for event, data in self.buffer:
            if event not in ("debug-info", "dirty-flag", "write-source-info"):
                rv.append((event, data))
        return rv

    def get_failures(self):
        rv = []
        for event, data in self.buffer:
            if event == "failure":
                rv.append(data)
        return rv

    def _emit(self, _event, **extra):
        self.buffer.append((_event, extra))

    def start_build(self, activity):
        self._emit("start-build", activity=activity)

    def finish_build(self, activity, start_time):
        self._emit("finish-build", activity=activity)

    def start_artifact_build(self, is_current):
        self._emit(
            "start-artifact-build",
            artifact=self.current_artifact,
            is_current=is_current,
        )

    def finish_artifact_build(self, start_time):
        self._emit("finish-artifact-build", artifact=self.current_artifact)

    def report_build_all_failure(self, failures):
        self._emit("build-all-failure", failures=failures)

    def report_failure(self, artifact, exc_info):
        self._emit("failure", artifact=artifact, exc_info=exc_info)

    def report_dirty_flag(self, value):
        self._emit("dirty-flag", artifact=self.current_artifact, value=value)

    def report_write_source_info(self, info):
        self._emit("write-source-info", info=info, artifact=self.current_artifact)

    def report_prune_source_info(self, source):
        self._emit("prune-source-info", source=source)

    def report_build_func(self, build_func):
        self._emit("build-func", func=describe_build_func(build_func))

    def report_sub_artifact(self, artifact):
        self._emit("sub-artifact", artifact=artifact)

    def report_debug_info(self, key, value):
        self._emit("debug-info", key=key, value=value)

    def report_generic(self, message):
        self._emit("generic", message=message)

    def enter_source(self):
        self._emit("enter-source", source=self.current_source)

    def leave_source(self, start_time):
        self._emit("leave-source", source=self.current_source)

    def report_pruned_artifact(self, artifact_name):
        self._emit("pruned-artifact", artifact_name=artifact_name)


class CliReporter(Reporter):
    def __init__(self, env, verbosity=0):
        Reporter.__init__(self, env, verbosity)
        self.indentation = 0

    def indent(self):
        self.indentation += 1

    def outdent(self):
        self.indentation -= 1

    def _write_line(self, text):
        click.echo(" " * (self.indentation * 2) + text)

    def _write_kv_info(self, key, value):
        self._write_line("%s: %s" % (key, style(str(value), fg="yellow")))

    def start_build(self, activity):
        self._write_line(style("Started %s" % activity, fg="cyan"))
        if not self.show_build_info:
            return
        self._write_line(style("  Tree: %s" % self.env.root_path, fg="cyan"))
        self._write_line(
            style("  Output path: %s" % self.builder.destination_path, fg="cyan")
        )

    def finish_build(self, activity, start_time):
        self._write_line(
            style(
                "Finished %s in %.2f sec" % (activity, time.time() - start_time),
                fg="cyan",
            )
        )

    def start_artifact_build(self, is_current):
        artifact = self.current_artifact
        if is_current:
            if not self.show_current_artifacts:
                return
            sign = click.style("X", fg="cyan")
        else:
            sign = click.style("U", fg="green")
        self._write_line("%s %s" % (sign, artifact.artifact_name))

        self.indent()

    def finish_artifact_build(self, start_time):
        self.outdent()

    def report_build_all_failure(self, failures):
        self._write_line(
            click.style(
                "Error: Build failed with %s failure%s."
                % (failures, failures != 1 and "s" or ""),
                fg="red",
            )
        )

    def report_failure(self, artifact, exc_info):
        sign = click.style("E", fg="red")
        err = " ".join(
            "".join(traceback.format_exception_only(*exc_info[:2])).splitlines()
        ).strip()
        self._write_line("%s %s (%s)" % (sign, artifact.artifact_name, err))

        if not self.show_tracebacks:
            return

        tb = traceback.format_exception(*exc_info)
        for line in "".join(tb).splitlines():
            if line.startswith("Traceback "):
                line = click.style(line, fg="red")
            elif line.startswith("  File "):
                line = click.style(line, fg="yellow")
            elif not line.startswith("    "):
                line = click.style(line, fg="red")
            self._write_line("  " + line)

    def report_dirty_flag(self, value):
        if self.show_artifact_internals and (value or self.show_debug_info):
            self._write_kv_info("forcing sources dirty", value)

    def report_write_source_info(self, info):
        if self.show_artifact_internals and self.show_debug_info:
            self._write_kv_info(
                "writing source info", "%s [%s]" % (info.title_i18n["en"], info.type)
            )

    def report_prune_source_info(self, source):
        if self.show_artifact_internals and self.show_debug_info:
            self._write_kv_info("pruning source info", source)

    def report_build_func(self, build_func):
        if self.show_artifact_internals:
            self._write_kv_info("build program", describe_build_func(build_func))

    def report_sub_artifact(self, artifact):
        if self.show_artifact_internals:
            self._write_kv_info("sub artifact", artifact.artifact_name)

    def report_debug_info(self, key, value):
        if self.show_debug_info:
            self._write_kv_info(key, value)

    def report_generic(self, message):
        self._write_line(style(str(message), fg="cyan"))

    def enter_source(self):
        if not self.show_source_internals:
            return
        self._write_line("Source %s" % style(repr(self.current_source), fg="magenta"))
        self.indent()

    def leave_source(self, start_time):
        if self.show_source_internals:
            self.outdent()

    def report_pruned_artifact(self, artifact_name):
        self._write_line("%s %s" % (style("D", fg="red"), artifact_name))


null_reporter = NullReporter(None)


@LocalProxy
def reporter():
    rv = _reporter_stack.top
    if rv is None:
        rv = null_reporter
    return rv



# -*- coding: utf-8 -*-
import codecs
import hashlib
import json
import multiprocessing
import os
import posixpath
import re
import subprocess
import sys
import tempfile
import traceback
import unicodedata
import uuid
from contextlib import contextmanager
from datetime import datetime
from functools import lru_cache
from pathlib import PurePosixPath
from queue import Queue
from threading import Thread

import click
from jinja2 import is_undefined
from markupsafe import Markup
from slugify import slugify as _slugify
from werkzeug import urls
from werkzeug.http import http_date
from werkzeug.urls import url_parse


is_windows = os.name == "nt"

_slash_escape = "\\/" not in json.dumps("/")

_slashes_re = re.compile(r"(/\.{1,2}(/|$))|/")
_last_num_re = re.compile(r"^(.*)(\d+)(.*?)$")
_list_marker = object()
_value_marker = object()

# Figure out our fs encoding, if it's ascii we upgrade to utf-8
fs_enc = sys.getfilesystemencoding()
try:
    if codecs.lookup(fs_enc).name == "ascii":
        fs_enc = "utf-8"
except LookupError:
    pass


def split_virtual_path(path):
    if "@" in path:
        return path.split("@", 1)
    return path, None


def _norm_join(a, b):
    return posixpath.normpath(posixpath.join(a, b))


def join_path(a, b):
    a_p, a_v = split_virtual_path(a)
    b_p, b_v = split_virtual_path(b)

    # Special case: paginations are considered special virtual paths
    # where the parent is the actual parent of the page.  This however
    # is explicitly not done if the path we join with refers to the
    # current path (empty string or dot).
    if b_p not in ("", ".") and a_v and a_v.isdigit():
        a_v = None

    # New path has a virtual path, add that to it.
    if b_v:
        rv = _norm_join(a_p, b_p) + "@" + b_v
    elif a_v:
        rv = a_p + "@" + _norm_join(a_v, b_p)
    else:
        rv = _norm_join(a_p, b_p)
    if rv[-2:] == "@.":
        rv = rv[:-2]
    return rv


def cleanup_path(path):
    return "/" + _slashes_re.sub("/", path).strip("/")


def parse_path(path):
    x = cleanup_path(path).strip("/").split("/")
    if x == [""]:
        return []
    return x


def is_path_child_of(a, b, strict=True):
    a_p, a_v = split_virtual_path(a)
    b_p, b_v = split_virtual_path(b)
    a_p = parse_path(a_p)
    b_p = parse_path(b_p)
    a_v = parse_path(a_v or "")
    b_v = parse_path(b_v or "")

    if not strict and a_p == b_p and a_v == b_v:
        return True
    if not a_v and b_v:
        return False
    if a_p == b_p and a_v[: len(b_v)] == b_v and len(a_v) > len(b_v):
        return True
    return a_p[: len(b_p)] == b_p and len(a_p) > len(b_p)


def untrusted_to_os_path(path):
    path = path.strip("/").replace("/", os.path.sep)
    if not isinstance(path, str):
        path = path.decode(fs_enc, "replace")
    return path


def is_path(path):
    return os.path.sep in path or (os.path.altsep and os.path.altsep in path)


def magic_split_ext(filename, ext_check=True):
    """Splits a filename into base and extension.  If ext check is enabled
    (which is the default) then it verifies the extension is at least
    reasonable.
    """

    def bad_ext(ext):
        if not ext_check:
            return False
        if not ext or ext.split() != [ext] or ext.strip() != ext:
            return True
        return False

    parts = filename.rsplit(".", 2)
    if len(parts) == 1:
        return parts[0], ""
    if len(parts) == 2 and not parts[0]:
        return "." + parts[1], ""
    if len(parts) == 3 and len(parts[1]) < 5:
        ext = ".".join(parts[1:])
        if not bad_ext(ext):
            return parts[0], ext
    ext = parts[-1]
    if bad_ext(ext):
        return filename, ""
    basename = ".".join(parts[:-1])
    return basename, ext


def iter_dotted_path_prefixes(dotted_path):
    pieces = dotted_path.split(".")
    if len(pieces) == 1:
        yield dotted_path, None
    else:
        for x in range(1, len(pieces)):
            yield ".".join(pieces[:x]), ".".join(pieces[x:])


def resolve_dotted_value(obj, dotted_path):
    node = obj
    for key in dotted_path.split("."):
        if isinstance(node, dict):
            new_node = node.get(key)
            if new_node is None and key.isdigit():
                new_node = node.get(int(key))
        elif isinstance(node, list):
            try:
                new_node = node[int(key)]
            except (ValueError, TypeError, IndexError):
                new_node = None
        else:
            new_node = None
        node = new_node
        if node is None:
            break
    return node


def decode_flat_data(itemiter, dict_cls=dict):
    def _split_key(name):
        result = name.split(".")
        for idx, part in enumerate(result):
            if part.isdigit():
                result[idx] = int(part)
        return result

    def _enter_container(container, key):
        if key not in container:
            return container.setdefault(key, dict_cls())
        return container[key]

    def _convert(container):
        if _value_marker in container:
            force_list = False
            values = container.pop(_value_marker)
            if container.pop(_list_marker, False):
                force_list = True
                values.extend(_convert(x[1]) for x in sorted(container.items()))
            if not force_list and len(values) == 1:
                values = values[0]

            if not container:
                return values
            return _convert(container)
        if container.pop(_list_marker, False):
            return [_convert(x[1]) for x in sorted(container.items())]
        return dict_cls((k, _convert(v)) for k, v in container.items())

    result = dict_cls()

    for key, value in itemiter:
        parts = _split_key(key)
        if not parts:
            continue
        container = result
        for part in parts:
            last_container = container
            container = _enter_container(container, part)
            last_container[_list_marker] = isinstance(part, int)
        container[_value_marker] = [value]

    return _convert(result)


def merge(a, b):
    """Merges two values together."""
    if b is None and a is not None:
        return a
    if a is None:
        return b
    if isinstance(a, list) and isinstance(b, list):
        for idx, (item_1, item_2) in enumerate(zip(a, b)):
            a[idx] = merge(item_1, item_2)
    if isinstance(a, dict) and isinstance(b, dict):
        for key, value in b.items():
            a[key] = merge(a.get(key), value)
        return a
    return a


def slugify(text):
    """
    A wrapper around python-slugify which preserves file extensions
    and forward slashes.
    """

    parts = text.split("/")
    parts[-1], ext = magic_split_ext(parts[-1])

    out = "/".join(_slugify(part) for part in parts)

    if ext:
        return out + "." + ext
    return out


def secure_filename(filename, fallback_name="file"):
    base = filename.replace("/", " ").replace("\\", " ")
    basename, ext = magic_split_ext(base)
    rv = slugify(basename).lstrip(".")
    if not rv:
        rv = fallback_name
    if ext:
        return rv + "." + ext
    return rv


def increment_filename(filename):
    directory, filename = os.path.split(filename)
    basename, ext = magic_split_ext(filename, ext_check=False)

    match = _last_num_re.match(basename)
    if match is not None:
        rv = match.group(1) + str(int(match.group(2)) + 1) + match.group(3)
    else:
        rv = basename + "2"

    if ext:
        rv += "." + ext
    if directory:
        return os.path.join(directory, rv)
    return rv


@lru_cache(maxsize=None)
def locate_executable(exe_file, cwd=None, include_bundle_path=True):
    """Locates an executable in the search path."""
    choices = [exe_file]
    resolve = True

    # If it's already a path, we don't resolve.
    if os.path.sep in exe_file or (os.path.altsep and os.path.altsep in exe_file):
        resolve = False

    extensions = os.environ.get("PATHEXT", "").split(";")
    _, ext = os.path.splitext(exe_file)
    if (
        os.name != "nt"
        and "" not in extensions
        or any(ext.lower() == extension.lower() for extension in extensions)
    ):
        extensions.insert(0, "")

    if resolve:
        paths = os.environ.get("PATH", "").split(os.pathsep)
        choices = [os.path.join(path, exe_file) for path in paths]

    if os.name == "nt":
        choices.append(os.path.join((cwd or os.getcwd()), exe_file))

    try:
        for path in choices:
            for ext in extensions:
                if os.access(path + ext, os.X_OK):
                    return path + ext
        return None
    except OSError:
        return None


class JSONEncoder(json.JSONEncoder):
    def default(self, o):  # pylint: disable=method-hidden
        if is_undefined(o):
            return None
        if isinstance(o, datetime):
            return http_date(o)
        if isinstance(o, uuid.UUID):
            return str(o)
        if hasattr(o, "__html__"):
            return str(o.__html__())
        return json.JSONEncoder.default(self, o)


def htmlsafe_json_dump(obj, **kwargs):
    kwargs.setdefault("cls", JSONEncoder)
    rv = (
        json.dumps(obj, **kwargs)
        .replace("<", "\\u003c")
        .replace(">", "\\u003e")
        .replace("&", "\\u0026")
        .replace("'", "\\u0027")
    )
    if not _slash_escape:
        rv = rv.replace("\\/", "/")
    return rv


def tojson_filter(obj, **kwargs):
    return Markup(htmlsafe_json_dump(obj, **kwargs))


def safe_call(func, args=None, kwargs=None):
    try:
        return func(*(args or ()), **(kwargs or {}))
    except Exception:
        # XXX: logging
        traceback.print_exc()
        return None


class Worker(Thread):
    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.daemon = True
        self.start()

    def run(self):
        while 1:
            func, args, kwargs = self.tasks.get()
            safe_call(func, args, kwargs)
            self.tasks.task_done()


class WorkerPool:
    def __init__(self, num_threads=None):
        if num_threads is None:
            num_threads = multiprocessing.cpu_count()
        self.tasks = Queue(num_threads)
        for _ in range(num_threads):
            Worker(self.tasks)

    def add_task(self, func, *args, **kargs):
        self.tasks.put((func, args, kargs))

    def wait_for_completion(self):
        self.tasks.join()


class Url:
    def __init__(self, value):
        self.url = value
        u = url_parse(value)
        i = u.to_iri_tuple()
        self.ascii_url = str(u)
        self.host = i.host
        self.ascii_host = u.ascii_host
        self.port = u.port
        self.path = i.path
        self.query = u.query
        self.anchor = i.fragment
        self.scheme = u.scheme

    def __unicode__(self):
        return self.url

    def __str__(self):
        return self.ascii_url


def is_unsafe_to_delete(path, base):
    a = os.path.abspath(path)
    b = os.path.abspath(base)
    diff = os.path.relpath(a, b)
    first = diff.split(os.path.sep)[0]
    return first in (os.path.curdir, os.path.pardir)


def prune_file_and_folder(name, base):
    if is_unsafe_to_delete(name, base):
        return False
    try:
        os.remove(name)
    except OSError:
        try:
            os.rmdir(name)
        except OSError:
            return False
    head, tail = os.path.split(name)
    if not tail:
        head, tail = os.path.split(head)
    while head and tail:
        try:
            if is_unsafe_to_delete(head, base):
                return False
            os.rmdir(head)
        except OSError:
            break
        head, tail = os.path.split(head)
    return True


def sort_normalize_string(s):
    return unicodedata.normalize("NFD", str(s).lower().strip())


def get_dependent_url(url_path, suffix, ext=None):
    url_directory, url_filename = posixpath.split(url_path)
    url_base, url_ext = posixpath.splitext(url_filename)
    if ext is None:
        ext = url_ext
    return posixpath.join(url_directory, url_base + "@" + suffix + ext)


@contextmanager
def atomic_open(filename, mode="r", encoding=None):
    if "r" not in mode:
        fd, tmp_filename = tempfile.mkstemp(
            dir=os.path.dirname(filename), prefix=".__atomic-write"
        )
        os.chmod(tmp_filename, 0o644)
        f = os.fdopen(fd, mode)
    else:
        f = open(filename, mode=mode, encoding=encoding)
        tmp_filename = None
    try:
        yield f
    except Exception as e:
        f.close()
        _exc_type, exc_value, tb = sys.exc_info()
        if tmp_filename is not None:
            try:
                os.remove(tmp_filename)
            except OSError:
                pass

        if exc_value.__traceback__ is not tb:
            raise exc_value.with_traceback(tb) from e
        raise exc_value from e
    else:
        f.close()
        if tmp_filename is not None:
            os.replace(tmp_filename, filename)


def portable_popen(cmd, *args, **kwargs):
    """A portable version of subprocess.Popen that automatically locates
    executables before invoking them.  This also looks for executables
    in the bundle bin.
    """
    if cmd[0] is None:
        raise RuntimeError("No executable specified")
    exe = locate_executable(cmd[0], kwargs.get("cwd"))
    if exe is None:
        raise RuntimeError('Could not locate executable "%s"' % cmd[0])

    if isinstance(exe, str) and sys.platform != "win32":
        exe = exe.encode(sys.getfilesystemencoding())
    cmd[0] = exe
    return subprocess.Popen(cmd, *args, **kwargs)


def is_valid_id(value):
    if value == "":
        return True
    return (
        "/" not in value
        and value.strip() == value
        and value.split() == [value]
        and not value.startswith(".")
    )


def secure_url(url):
    url = urls.url_parse(url)
    if url.password is not None:
        url = url.replace(
            netloc="%s@%s"
            % (
                url.username,
                url.netloc.split("@")[-1],
            )
        )
    return url.to_url()


def bool_from_string(val, default=None):
    if val in (True, False, 1, 0):
        return bool(val)
    if isinstance(val, str):
        val = val.lower()
        if val in ("true", "yes", "1"):
            return True
        if val in ("false", "no", "0"):
            return False
    return default


def make_relative_url(source, target):
    """
    Returns the relative path (url) needed to navigate
    from `source` to `target`.
    """

    # WARNING: this logic makes some unwarranted assumptions about
    # what is a directory and what isn't. Ideally, this function
    # would be aware of the actual filesystem.
    s_is_dir = source.endswith("/")
    t_is_dir = target.endswith("/")

    source = PurePosixPath(posixpath.normpath(source))
    target = PurePosixPath(posixpath.normpath(target))

    if not s_is_dir:
        source = source.parent

    relpath = str(get_relative_path(source, target))
    if t_is_dir:
        relpath += "/"

    return relpath


def get_relative_path(source, target):
    """
    Returns the relative path needed to navigate from `source` to `target`.

    get_relative_path(source: PurePosixPath,
                      target: PurePosixPath) -> PurePosixPath
    """

    if not source.is_absolute() and target.is_absolute():
        raise ValueError("Cannot navigate from a relative path" " to an absolute one")

    if source.is_absolute() and not target.is_absolute():
        # nothing to do
        return target

    if source.is_absolute() and target.is_absolute():
        # convert them to relative paths to simplify the logic
        source = source.relative_to("/")
        target = target.relative_to("/")

    # is the source an ancestor of the target?
    try:
        return target.relative_to(source)
    except ValueError:
        pass

    # even if it isn't, one of the source's ancestors might be
    # (and if not, the root will be the common ancestor)
    distance = PurePosixPath(".")
    for ancestor in source.parents:
        distance /= ".."

        try:
            relpath = target.relative_to(ancestor)
        except ValueError:
            continue
        else:
            # prepend the distance to the common ancestor
            return distance / relpath
    # We should never get here.  (The last ancestor in source.parents will
    # be '.' — target.relative_to('.') will always succeed.)
    raise AssertionError("This should not happen")


def get_structure_hash(params):
    """Given a Python structure this generates a hash.  This is useful for
    storing artifact config hashes.  Not all Python types are supported, but
    quite a few are.
    """
    h = hashlib.md5()

    def _hash(obj):
        if obj is None:
            h.update("N;")
        elif obj is True:
            h.update("T;")
        elif obj is False:
            h.update("F;")
        elif isinstance(obj, dict):
            h.update("D%d;" % len(obj))
            for key, value in sorted(obj.items()):
                _hash(key)
                _hash(value)
        elif isinstance(obj, tuple):
            h.update("T%d;" % len(obj))
            for item in obj:
                _hash(item)
        elif isinstance(obj, list):
            h.update("L%d;" % len(obj))
            for item in obj:
                _hash(item)
        elif isinstance(obj, int):
            h.update("T%d;" % obj)
        elif isinstance(obj, bytes):
            h.update("B%d;%s;" % (len(obj), obj))
        elif isinstance(obj, str):
            h.update("S%d;%s;" % (len(obj), obj.encode("utf-8")))
        elif hasattr(obj, "__get_lektor_param_hash__"):
            obj.__get_lektor_param_hash__(h)

    _hash(params)
    return h.hexdigest()


def profile_func(func):
    # pylint: disable=import-outside-toplevel

    from cProfile import Profile
    from pstats import Stats

    p = Profile()
    rv = []
    p.runcall(lambda: rv.append(func()))
    p.dump_stats("/tmp/lektor-%s.prof" % func.__name__)

    stats = Stats(p, stream=sys.stderr)
    stats.sort_stats("time", "calls")
    stats.print_stats()

    return rv[0]


def deg_to_dms(deg):
    d = int(deg)
    md = abs(deg - d) * 60
    m = int(md)
    sd = (md - m) * 60
    return (d, m, sd)


def format_lat_long(lat=None, long=None, secs=True):
    def _format(value, sign):
        d, m, sd = deg_to_dms(value)
        return "%d° %d′ %s%s" % (
            abs(d),
            abs(m),
            secs and ("%d″ " % abs(sd)) or "",
            sign[d < 0],
        )

    rv = []
    if lat is not None:
        rv.append(_format(lat, "NS"))
    if long is not None:
        rv.append(_format(long, "EW"))
    return ", ".join(rv)


def get_app_dir():
    return click.get_app_dir("Lektor")


def get_cache_dir():
    if is_windows:
        folder = os.environ.get("LOCALAPPDATA")
        if folder is None:
            folder = os.environ.get("APPDATA")
            if folder is None:
                folder = os.path.expanduser("~")
        return os.path.join(folder, "Lektor", "Cache")
    if sys.platform == "darwin":
        return os.path.join(os.path.expanduser("~/Library/Caches/Lektor"))
    return os.path.join(
        os.environ.get("XDG_CACHE_HOME", os.path.expanduser("~/.cache")), "lektor"
    )


class URLBuilder:
    def __init__(self):
        self.items = []

    def append(self, item):
        if item is None:
            return
        item = str(item).strip("/")
        if item:
            self.items.append(item)

    def get_url(self, trailing_slash=None):
        url = "/" + "/".join(self.items)
        if trailing_slash is not None and not trailing_slash:
            return url
        if url == "/":
            return url
        if trailing_slash is None:
            _, last = url.split("/", 1)
            if "." in last:
                return url
        return url + "/"


def build_url(iterable, trailing_slash=None):
    builder = URLBuilder()
    for item in iterable:
        builder.append(item)
    return builder.get_url(trailing_slash=trailing_slash)


def comma_delimited(s):
    """Split a comma-delimited string."""
    for part in s.split(","):
        stripped = part.strip()
        if stripped:
            yield stripped


def process_extra_flags(flags):
    if isinstance(flags, dict):
        return flags
    rv = {}
    for flag in flags or ():
        if ":" in flag:
            k, v = flag.split(":", 1)
            rv[k] = v
        else:
            rv[flag] = flag
    return rv



class Pagination:
    def __init__(self, record, pagination_config):
        #: the pagination config
        self.config = pagination_config
        #: the current page's record
        self.current = record
        #: the current page number (1 indexed)
        self.page = record.page_num
        #: the number of items to be displayed on a page.
        self.per_page = pagination_config.per_page
        #: the total number of items matching the query
        self.total = pagination_config.count_total_items(record)

    @property
    def items(self):
        """The children for this page."""
        return self.config.slice_query_for_page(self.current, self.page)

    @property
    def pages(self):
        """The total number of pages."""
        pages = (self.total + self.per_page - 1) // self.per_page
        # Even when there are no children, we want at least one page
        return max(pages, 1)

    @property
    def prev_num(self):
        """The page number of the previous page."""
        if self.page > 1:
            return self.page - 1
        return None

    @property
    def has_prev(self):
        """True if a previous page exists."""
        return self.page > 1

    @property
    def prev(self):
        """The record for the previous page."""
        if not self.has_prev:
            return None
        return self.config.get_record_for_page(self.current, self.page - 1)

    @property
    def has_next(self):
        """True if a following page exists."""
        return self.page < self.pages

    @property
    def next_num(self):
        """The page number of the following page."""
        if self.page < self.pages:
            return self.page + 1
        return None

    @property
    def next(self):
        """The record for the following page."""
        if not self.has_next:
            return None
        return self.config.get_record_for_page(self.current, self.page + 1)

    def for_page(self, page):
        """Returns the record for a specific page."""
        if 1 <= page <= self.pages:
            return self.config.get_record_for_page(self.current, page)
        return None

    def iter_pages(self, left_edge=2, left_current=2, right_current=5, right_edge=2):
        """Iterate over the page numbers in the pagination, with elision.

        In the general case, this returns the concatenation of three ranges:

            1. A range (always starting at page one) at the beginning
               of the page number sequence.  The length of the this
               range is specified by the ``left_edge`` argument (which
               may be zero).

            2. A range around the current page.  This range will
               include ``left_current`` pages before, and
               ``right_current`` pages after the current page.  This
               range always includes the current page.

            3. Finally, a range (always ending at the last page) at
               the end of the page sequence.  The length of this range
               is specified by the ``right_edge`` argument.

        If any of these ranges overlap, they will be merged.  A
        ``None`` will be inserted between non-overlapping ranges to
        signify that pages have been elided.

        This is how you could render such a pagination in the templates:
        .. sourcecode:: html+jinja
            {% macro render_pagination(pagination, endpoint) %}
              <div class=pagination>
              {%- for page in pagination.iter_pages() %}
                {% if page %}
                  {% if page != pagination.page %}
                    <a href="{{ url_for(endpoint, page=page) }}">{{ page }}</a>
                  {% else %}
                    <strong>{{ page }}</strong>
                  {% endif %}
                {% else %}
                  <span class=ellipsis>...</span>
                {% endif %}
              {%- endfor %}
              </div>
            {% endmacro %}

        """
        last = 0
        for num in range(1, self.pages + 1):
            # pylint: disable=chained-comparison
            if (
                num <= left_edge
                or (
                    num >= self.page - left_current and num <= self.page + right_current
                )
                or num > self.pages - right_edge
            ):
                if last + 1 != num:
                    yield None
                yield num
                last = num
        if last != self.pages:
            yield None



from contextlib import contextmanager

from jinja2 import Undefined
from werkzeug.local import LocalProxy
from werkzeug.local import LocalStack

from lektor.reporter import reporter


_ctx_stack = LocalStack()


def url_to(*args, **kwargs):
    """Calculates a URL to another record."""
    ctx = get_ctx()
    if ctx is None:
        raise RuntimeError("No context found")
    return ctx.url_to(*args, **kwargs)


def get_asset_url(asset):
    """Calculates the asset URL relative to the current record."""
    ctx = get_ctx()
    if ctx is None:
        raise RuntimeError("No context found")
    asset = site_proxy.get_asset(asset)
    if asset is None:
        return Undefined("Asset not found")
    info = ctx.build_state.get_file_info(asset.source_filename)
    return "%s?h=%s" % (
        ctx.source.url_to("!" + asset.url_path),
        info.checksum[:8],
    )


@LocalProxy
def site_proxy():
    """Returns the current pad."""
    ctx = get_ctx()
    if ctx is None:
        return Undefined(hint="Cannot access the site from here", name="site")
    return ctx.pad


@LocalProxy
def config_proxy():
    """Returns the current config."""
    return site_proxy.db.config


def get_ctx():
    """Returns the current context."""
    return _ctx_stack.top


def get_locale(default="en_US"):
    """Returns the current locale."""
    ctx = get_ctx()
    if ctx is not None:
        rv = ctx.locale
        if rv is not None:
            return rv
        return ctx.pad.db.config.site_locale
    return default


class Context:
    """The context is a thread local object that provides the system with
    general information about in which state it is.  The context is created
    whenever a source is processed and can be accessed by template engine and
    other things.

    It's considered read and write and also accumulates changes that happen
    during processing of the object.
    """

    def __init__(self, artifact=None, pad=None):
        if pad is None:
            if artifact is None:
                raise TypeError(
                    "Either artifact or pad is needed to " "construct a context."
                )
            pad = artifact.build_state.pad

        if artifact is not None:
            self.artifact = artifact
            self.source = artifact.source_obj
            self.build_state = self.artifact.build_state
        else:
            self.artifact = None
            self.source = None
            self.build_state = None

        self.exc_info = None

        self.pad = pad

        # Processing information
        self.referenced_dependencies = set()
        self.referenced_virtual_dependencies = {}
        self.sub_artifacts = []

        self.flow_block_render_stack = []

        self._forced_base_url = None

        # General cache system where other things can put their temporary
        # stuff in.
        self.cache = {}

        self._dependency_collectors = []

    @property
    def env(self):
        """The environment of the context."""
        return self.pad.db.env

    @property
    def record(self):
        """If the source is a record it will be available here."""
        rv = self.source
        if rv is not None and rv.source_classification == "record":
            return rv
        return None

    @property
    def locale(self):
        """Returns the current locale if it's available, otherwise `None`.
        This does not fall back to the site locale.
        """
        source = self.source
        if source is not None:
            alt_cfg = self.pad.db.config["ALTERNATIVES"].get(source.alt)
            if alt_cfg:
                return alt_cfg["locale"]
        return None

    def push(self):
        _ctx_stack.push(self)

    @staticmethod
    def pop():
        _ctx_stack.pop()

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.pop()

    @property
    def base_url(self):
        """The URL path for the current context."""
        if self._forced_base_url:
            return self._forced_base_url
        if self.source is not None:
            return self.source.url_path
        return "/"

    def url_to(self, path, alt=None, absolute=None, external=None):
        """Returns a URL to another path."""
        if self.source is None:
            raise RuntimeError(
                "Can only generate paths to other pages if "
                "the context has a source document set."
            )
        rv = self.source.url_to(path, alt=alt, absolute=True)
        return self.pad.make_url(rv, self.base_url, absolute, external)

    def sub_artifact(self, *args, **kwargs):
        """Decorator version of :func:`add_sub_artifact`."""

        def decorator(f):
            self.add_sub_artifact(build_func=f, *args, **kwargs)
            return f

        return decorator

    def add_sub_artifact(
        self,
        artifact_name,
        build_func=None,
        sources=None,
        source_obj=None,
        config_hash=None,
    ):
        """Sometimes it can happen that while building an artifact another
        artifact needs building.  This function is generally used to record
        this request.
        """
        if self.build_state is None:
            raise TypeError(
                "The context does not have a build state which "
                "means that artifact declaration is not possible."
            )
        aft = self.build_state.new_artifact(
            artifact_name=artifact_name,
            sources=sources,
            source_obj=source_obj,
            config_hash=config_hash,
        )
        self.sub_artifacts.append((aft, build_func))
        reporter.report_sub_artifact(aft)

    def record_dependency(self, filename):
        """Records a dependency from processing."""
        self.referenced_dependencies.add(filename)
        for coll in self._dependency_collectors:
            coll(filename)

    def record_virtual_dependency(self, virtual_source):
        """Records a dependency from processing."""
        path = virtual_source.path
        self.referenced_virtual_dependencies[path] = virtual_source
        for coll in self._dependency_collectors:
            coll(virtual_source)

    @contextmanager
    def gather_dependencies(self, func):
        """For the duration of the `with` block the provided function will be
        invoked for all dependencies encountered.
        """
        self._dependency_collectors.append(func)
        try:
            yield
        finally:
            self._dependency_collectors.pop()

    @contextmanager
    def changed_base_url(self, value):
        """Temporarily overrides the URL path of the context."""
        old = self._forced_base_url
        self._forced_base_url = value
        try:
            yield
        finally:
            self._forced_base_url = old



import os
import re
import warnings

import pytest
from markers import imagemagick

from lektor.cli import cli
from lektor.project import Project


def test_build_abort_in_existing_nonempty_dir(project_cli_runner):
    os.mkdir("build_dir")
    with open("build_dir/test", "w", encoding="utf-8"):
        pass
    result = project_cli_runner.invoke(cli, ["build", "-O", "build_dir"], input="n\n")
    assert "Aborted!" in result.output
    assert result.exit_code == 1


@imagemagick
def test_build_continue_in_existing_nonempty_dir(project_cli_runner):
    os.mkdir("build_dir")
    with open("build_dir/test", "w", encoding="utf-8"):
        pass
    result = project_cli_runner.invoke(cli, ["build", "-O", "build_dir"], input="y\n")
    assert "Finished prune" in result.output
    assert result.exit_code == 0


def test_alias(project_cli_runner):
    result = project_cli_runner.invoke(cli, ["pr"])  # short for 'project-info'
    assert result.exit_code == 0
    assert "Name: Demo Project" in result.output


def test_dev_cmd_alias(isolated_cli_runner):
    result = isolated_cli_runner.invoke(cli, ["dev", "p"])  # short for 'publish-plugin'
    assert result.exit_code == 2
    assert "Error: This command must be run in a Lektor plugin folder" in result.output


def test_alias_multiple_matches(project_cli_runner):
    result = project_cli_runner.invoke(
        cli, ["p"]
    )  # short for 'project-info' & 'plugins'
    assert result.exit_code == 2
    assert "Error: Too many matches" in result.output


def test_alias_no_matches(project_cli_runner):
    result = project_cli_runner.invoke(cli, ["z"])
    assert result.exit_code == 2
    assert "Error: No such command" in result.output


def test_build_no_project(isolated_cli_runner):
    result = isolated_cli_runner.invoke(cli, ["build"])
    assert result.exit_code == 2
    assert "Could not automatically discover project." in result.output


@imagemagick
def test_build(project_cli_runner):
    result = project_cli_runner.invoke(cli, ["build"])
    assert (
        "files or folders already exist" not in result.output
    )  # No warning on fresh build
    assert result.exit_code == 0
    start_matches = re.findall(r"Started build", result.output)
    assert len(start_matches) == 1
    finish_matches = re.findall(r"Finished build in \d+\.\d{2} sec", result.output)
    assert len(finish_matches) == 1

    # rebuild
    result = project_cli_runner.invoke(cli, ["build"])
    assert (
        "files or folders already exist" not in result.output
    )  # No warning on repeat build
    assert result.exit_code == 0


def test_build_extra_flag(project_cli_runner, mocker):
    mock_builder = mocker.patch("lektor.builder.Builder")
    mock_builder.return_value.build_all.return_value = 0
    result = project_cli_runner.invoke(cli, ["build", "-f", "webpack"])
    assert result.exit_code == 0
    assert mock_builder.call_args[1]["extra_flags"] == ("webpack",)
    assert "use --extra-flag instead of --build-flag" not in result.output


def test_deprecated_build_flag(project_cli_runner, mocker):
    mock_builder = mocker.patch("lektor.builder.Builder")
    mock_builder.return_value.build_all.return_value = 0
    with warnings.catch_warnings(record=True) as w:
        result = project_cli_runner.invoke(cli, ["build", "--build-flag", "webpack"])
        assert result.exit_code == 0
        assert mock_builder.call_args[1]["extra_flags"] == ("webpack",)
        assert w
        assert "use --extra-flag instead of --build-flag" in str(w[0].message)


def test_deploy_extra_flag(project_cli_runner, mocker):
    mock_publish = mocker.patch("lektor.publisher.publish")
    result = project_cli_runner.invoke(cli, ["deploy", "-f", "draft"])
    assert result.exit_code == 0
    assert mock_publish.call_args[1]["extra_flags"] == ("draft",)


@pytest.mark.parametrize(
    "flag, expect",
    [
        ("--name", "Demo Project"),
        ("--project-file", "{tree_dir}{os.sep}Website.lektorproject"),
        ("--tree", "{tree_dir}"),
        ("--output-path", "{output_path}"),
    ],
)
def test_project_info_path_flags(project_cli_runner, flag, expect):
    tree_dir = os.getcwd()
    result = project_cli_runner.invoke(cli, ["project-info", flag])
    assert result.exit_code == 0
    assert result.stdout.rstrip() == expect.format(
        tree_dir=tree_dir,
        output_path=Project.from_path(tree_dir).get_output_path(),
        os=os,
    )



import os
import queue
import time
from collections import OrderedDict
from itertools import zip_longest

import click
from watchdog.events import DirModifiedEvent
from watchdog.events import FileMovedEvent
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.polling import PollingObserver

from lektor.utils import get_cache_dir


class EventHandler(FileSystemEventHandler):
    def __init__(self):
        super().__init__()
        self.queue = queue.Queue()

    def on_any_event(self, event):
        if not isinstance(event, DirModifiedEvent):
            if isinstance(event, FileMovedEvent):
                path = event.dest_path
            else:
                path = event.src_path
            self.queue.put((time.time(), event.event_type, path))


def _fullname(cls):
    """Return the full name of a class (including the module name)."""
    return f"{cls.__module__}.{cls.__qualname__}"


def _unique_everseen(seq):
    """Return the unique elements in sequence, preserving order."""
    return OrderedDict.fromkeys(seq).keys()


class BasicWatcher:
    def __init__(self, paths, observer_classes=(Observer, PollingObserver)):
        self.event_handler = EventHandler()
        self.paths = paths
        self.observer_classes = observer_classes
        self.observer = None

    def start(self):
        # Remove duplicates since there is no point in trying a given
        # observer class more than once. (This also simplifies the logic
        # for presenting sensible warning messages about broken
        # observers.)
        observer_classes = list(_unique_everseen(self.observer_classes))
        for observer_class, next_observer_class in zip_longest(
            observer_classes, observer_classes[1:]
        ):
            try:
                self._start_observer(observer_class)
                return
            except Exception as exc:
                if next_observer_class is None:
                    raise
                click.secho(
                    f"Creation of {_fullname(observer_class)} failed with exception:\n"
                    f"  {exc.__class__.__name__}: {exc!s}\n"
                    "This may be due to a configuration or other issue with your system.\n"
                    f"Falling back to {_fullname(next_observer_class)}.",
                    fg="red",
                    bold=True,
                )

    def stop(self):
        self.observer.stop()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, ex_type, ex_value, ex_tb):
        self.stop()

    def _start_observer(self, observer_class=Observer):
        if self.observer is not None:
            raise RuntimeError("Watcher already started.")
        observer = observer_class()
        for path in self.paths:
            observer.schedule(self.event_handler, path, recursive=True)
        observer.daemon = True
        observer.start()
        self.observer = observer

    def is_interesting(self, time, event_type, path):
        # pylint: disable=no-self-use
        return True

    def __iter__(self):
        while 1:
            time_, event_type, path = self.event_handler.queue.get()
            if self.is_interesting(time_, event_type, path):
                yield time_, event_type, path


class Watcher(BasicWatcher):
    def __init__(self, env, output_path=None):
        BasicWatcher.__init__(self, paths=[env.root_path] + env.theme_paths)
        self.env = env
        self.output_path = output_path
        self.cache_dir = os.path.abspath(get_cache_dir())

    def is_interesting(self, time, event_type, path):
        path = os.path.abspath(path)

        if self.env.is_uninteresting_source_name(os.path.basename(path)):
            return False
        if path.startswith(self.cache_dir):
            return False
        if self.output_path is not None and path.startswith(
            os.path.abspath(self.output_path)
        ):
            return False
        return True


def watch(env):
    """Returns a generator of file system events in the environment."""
    with Watcher(env) as watcher:
        try:
            for event in watcher:
                yield event
        except KeyboardInterrupt:
            pass



from lektor.markdown import Markdown
from lektor.types.base import Type


class MarkdownDescriptor:
    def __init__(self, source):
        self.source = source

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        return Markdown(self.source, record=obj)


class MarkdownType(Type):
    widget = "multiline-text"

    def value_from_raw(self, raw):
        return MarkdownDescriptor(raw.value or "")



# coding: utf-8
import os

from lektor.build_programs import BuildError
from lektor.builder import Builder
from lektor.db import Database
from lektor.environment import Environment
from lektor.project import Project
from lektor.reporter import BufferReporter


def get_unicode_builder(tmpdir):
    proj = Project.from_path(os.path.join(os.path.dirname(__file__), "ünicöde-project"))
    env = Environment(proj)
    pad = Database(env).new_pad()

    return pad, Builder(pad, str(tmpdir.mkdir("output")))


def test_unicode_project_folder(tmpdir):
    pad, builder = get_unicode_builder(tmpdir)
    prog, _ = builder.build(pad.root)
    with prog.artifacts[0].open("rb") as f:
        assert f.read() == b"<h1>Hello</h1>\n<p>W\xc3\xb6rld</p>\n\n"


def test_unicode_attachment_filename(tmpdir):
    pad, builder = get_unicode_builder(tmpdir)

    with BufferReporter(builder.env) as reporter:
        prog, _ = builder.build(pad.root.attachments.first())

        failures = reporter.get_failures()
        assert len(failures) == 0

        with prog.artifacts[0].open("rb") as f:
            assert f.read().rstrip() == b"attachment"


def test_bad_file_ignored(tmpdir):
    pad, builder = get_unicode_builder(tmpdir)
    record = pad.root.children.first()
    with BufferReporter(builder.env) as reporter:
        builder.build(record)
        failures = reporter.get_failures()
        assert len(failures) == 1
        exc_info = failures[0]["exc_info"]
        assert exc_info[0] is BuildError
        assert (
            "The URL for this record contains non "
            "ASCII characters" in exc_info[1].message
        )



# pylint: disable=import-outside-toplevel
import json
import os
import warnings

import click

from lektor.i18n import get_default_lang
from lektor.i18n import is_valid_language
from lektor.project import Project


def echo_json(data):
    click.echo(json.dumps(data, indent=2).rstrip())


def pruneflag(cli):
    return click.option(
        "--prune/--no-prune",
        default=True,
        help="Controls if old " 'artifacts should be pruned.  "prune" is the default.',
    )(cli)


def extraflag(cli):
    return click.option(
        "-f",
        "--extra-flag",
        "extra_flags",
        multiple=True,
        help="Defines an arbitrary flag.  These can be used by plugins "
        "to customize the build and deploy process.  More information can be "
        "found in the documentation of affected plugins.",
    )(cli)


def _buildflag_deprecated(ctx, param, value):
    if value:
        warnings.warn(
            "use --extra-flag instead of --build-flag",
            DeprecationWarning,
        )
    return value


def buildflag(cli):
    return click.option(
        "--build-flag",
        "build_flags",
        multiple=True,
        help="Deprecated. Use --extra-flag instead.",
        callback=_buildflag_deprecated,
    )(cli)


class AliasedGroup(click.Group):

    # pylint: disable=inconsistent-return-statements
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        if len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail("Too many matches: %s" % ", ".join(sorted(matches)))


class Context:
    def __init__(self):
        self._project_path = os.environ.get("LEKTOR_PROJECT") or None
        self._project = None
        self._env = None
        self._ui_lang = None

    def _get_ui_lang(self):
        rv = self._ui_lang
        if rv is None:
            rv = self._ui_lang = get_default_lang()
        return rv

    def _set_ui_lang(self, value):
        self._ui_lang = value

    ui_lang = property(_get_ui_lang, _set_ui_lang)
    del _get_ui_lang, _set_ui_lang

    def set_project_path(self, value):
        self._project_path = value
        self._project = None

    def get_project(self, silent=False):
        if self._project is not None:
            return self._project
        if self._project_path is not None:
            rv = Project.from_path(self._project_path)
        else:
            rv = Project.discover()
        if rv is None:
            if silent:
                return None
            if self._project_path is None:
                raise click.UsageError(
                    "Could not automatically discover "
                    "project.  A Lektor project must "
                    "exist in the working directory or "
                    "any of the parent directories."
                )
            raise click.UsageError('Could not find project "%s"' % self._project_path)
        self._project = rv
        return rv

    def get_default_output_path(self):
        rv = os.environ.get("LEKTOR_OUTPUT_PATH")
        if rv is not None:
            return rv
        return self.get_project().get_output_path()

    def get_env(self, extra_flags=None):
        if self._env is not None:
            return self._env
        from lektor.environment import Environment

        env = Environment(
            self.get_project(), load_plugins=False, extra_flags=extra_flags
        )
        self._env = env
        return env

    def load_plugins(self, reinstall=False, extra_flags=None):
        from .packages import load_packages

        load_packages(self.get_env(extra_flags=extra_flags), reinstall=reinstall)

        if not reinstall:
            from .pluginsystem import initialize_plugins

            initialize_plugins(self.get_env())


pass_context = click.make_pass_decorator(Context, ensure=True)


def validate_language(ctx, param, value):
    if value is not None and not is_valid_language(value):
        raise click.BadParameter('Unsupported language "%s".' % value)
    return value



import functools
import queue
import threading

import py
import pytest
from watchdog.observers.api import BaseObserver
from watchdog.observers.polling import PollingObserver

from lektor import utils
from lektor.watcher import BasicWatcher
from lektor.watcher import watch
from lektor.watcher import Watcher


class IterateInThread(threading.Thread):
    """Iterate iterable in a separate thread.

    This is a iterator which yields the results of the iterable.
    I.e. mostly, it's transparent.

    The results are passed back to the calling thread via a queue.  If
    the queue remains empty for a bit of time a `Timeout` exception
    will be raised.

    The timeout prevents the test from getting hung when expected
    results are not forthcoming.

    """

    timeout = 2.0

    def __init__(self, it):
        threading.Thread.__init__(self, daemon=True)
        self.it = it
        self.queue = queue.Queue()
        self.start()

    def run(self):
        for item in self.it:
            self.queue.put(item)

    def __next__(self):
        try:
            return self.queue.get(timeout=self.timeout)
        except queue.Empty:
            return pytest.fail("Timed out waiting for iterator")


class BrokenObserver(PollingObserver):
    # The InotifyObserver, when it fails due to insufficient system
    # inotify resources, does not fail until an attempt is made to start it.
    def start(self):
        raise OSError("crapout")


class TestBasicWatcher:
    # pylint: disable=no-self-use

    @pytest.fixture
    def paths(self, tmp_path):
        return [str(tmp_path)]

    def test_creates_observer(self, paths):
        with BasicWatcher(paths) as watcher:
            assert isinstance(watcher.observer, BaseObserver)

    def test_default_observer_broken(self, paths, capsys):
        observer_classes = (BrokenObserver, PollingObserver)
        with BasicWatcher(paths, observer_classes=observer_classes) as watcher:
            assert watcher.observer.__class__ is PollingObserver
        assert "crapout" in capsys.readouterr().out

    def test_default_observer_is_polling(self, paths, capsys):
        observer_classes = (BrokenObserver, BrokenObserver)
        with pytest.raises(OSError, match=r"crapout"):
            with BasicWatcher(paths, observer_classes=observer_classes):
                pass
        assert capsys.readouterr() == ("", "")

    def test_perverse_usage(self, paths):
        # This exercises a bug which occurred when BasicWatcher was
        # called with repeated (failing) values in observer_classes.
        observer_classes = (BrokenObserver, BrokenObserver, PollingObserver)
        with BasicWatcher(paths, observer_classes=observer_classes) as watcher:
            assert isinstance(watcher.observer, BaseObserver)

    def test_raises_error_if_started_twice(self, paths):
        with BasicWatcher(paths) as watcher:
            with pytest.raises(RuntimeError, match="already started"):
                watcher.start()

    def test_iter(self, tmp_path):
        file1 = tmp_path / "file1"
        file1.touch()

        with BasicWatcher([str(tmp_path)]) as watcher:
            it = IterateInThread(watcher)

            file2 = tmp_path / "file2"
            file2.touch()
            # Check that we get notified about file2
            _, event_type, path = next(it)
            print(event_type, path)
            while path != str(file2):
                # On MacOS, for whatever reason, we get events about
                # the creation of tmp_path and file1.  Skip them.
                _, event_type, path = next(it)
                print(event_type, path)

            file1_renamed = tmp_path / "file1_renamed"
            file1.rename(file1_renamed)
            # Check for notification of renamed file.
            while path != str(file1_renamed):
                # Depending on platform, we may get more than one
                # event for file1. (E.g. on Linux we get both a
                # 'created' and a 'closed' event.)
                # (Also, on MacOS, for reasons not understood,
                # we appear to get a 'created' event for file1.)
                _, event_type, path = next(it)
                print(event_type, path)

            assert it.queue.empty()


def test_is_interesting(env):
    # pylint: disable=no-member
    cache_dir = py.path.local(utils.get_cache_dir())
    build_dir = py.path.local("build")

    w = Watcher(env, str(build_dir))

    # This partial makes the testing code shorter
    is_interesting = functools.partial(w.is_interesting, 0, "generic")

    assert is_interesting("a.file")
    assert not is_interesting(".file")
    assert not is_interesting(str(cache_dir / "another.file"))
    assert not is_interesting(str(build_dir / "output.file"))

    w.output_path = None
    assert is_interesting(str(build_dir / "output.file"))


def test_watch(env, mocker):
    Watcher = mocker.patch("lektor.watcher.Watcher")
    event1 = mocker.sentinel.event1

    def events():
        yield event1
        raise KeyboardInterrupt()

    watcher = Watcher.return_value.__enter__.return_value
    watcher.__iter__.return_value = events()

    assert list(watch(env)) == [event1]
